"""
Copyright (c) 2017, 2018, 2019 Irstea
Copyright (c) 2017, 2018, 2019 François Kneib
Copyright (c) 2017, 2018, 2019 Franck Bourrier
Copyright (c) 2017, 2018, 2019 David Toe
Copyright (c) 2017, 2018, 2019 Frédéric Berger
Copyright (c) 2017, 2018, 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""

"""
This module is used by the ThreeD model. It is a kind of master module that handles simulations.
"""

from . import Objects, Postprocessings, Engines
import platrock.Common.Debug as Debug
import platrock.Common.BounceModels as BounceModels
import platrock.Common.Math as Math
import platrock.Common.Simulations
import platrock.Common.ColorPalettes as cp
import platrock.Common.Outputs as Out
import numpy as np
import quaternion, time, sys
import osgeo.gdal as gdal
if(int(gdal.__version__.split('.')[0])<2):
	Debug.error("python3-gdal version is ",gdal.__version__,"but must be >=2. If you are using ubuntu 16.04 or earlier, you may consider using 'sudo add-apt-repository -y ppa:ubuntugis/ppa'.")
	sys.exit()
import shapely.geometry
import shapely.affinity
from matplotlib import cm
import copy,os,zipfile

#:The current simulation.
current_simulation=None

class Simulation(platrock.Common.Simulations.Simulation):
	"""
	A simulation
	
	.. _input_threeD_params:
	
	The table below displays the attributes that have to be set to the polygons of the input shapefile rocks_start_params_shapefile depending on the sub-model choosen.
	
	+------------------+--------------+---------+--------------+-----------+--------------+-------------+--------------+
	|                  |:attr:`number`|:attr:`z`|:attr:`volume`| :attr:`vz`|:attr:`length`|:attr:`width`|:attr:`height`|
	+------------------+--------------+---------+--------------+-----------+--------------+-------------+--------------+
	|PlatRock (builtin)|            * |       * |     *        |     *     |              |             |              |
	+------------------+--------------+---------+--------------+-----------+--------------+-------------+--------------+
	|Siconos           |            * |   *     |              |     *     |       *      |       *     |       *      |
	+------------------+--------------+---------+--------------+-----------+--------------+-------------+--------------+

	
	Args:
		rocks (list [:class:`ThreeD.Objects.Rock`]): list of all rocks, created in the launching script (see Examples/3D.py)
		current_rock (:class:`ThreeD.Objects.Rock`): the rock currently falling
		terrain (:class:`ThreeD.Objects.Terrain`): the terrain of the simulation
		gravity (float): the --- positive --- gravity acceleration value
		enable_forest (bool): whether to take trees into account or not
		engines (list [:class:`ThreeD.Engines.Engine`]): list of engines to run at each timestep
		dt (int): the time-step
		iter (int): the current iteration (reseted at each new rock)
		running (bool): whether the simulation is running
		enable_GUI (bool): enables the 3D view (experimental)
	"""
	webui_typename="PlatRock 3D"
	valid_input_shapefile_attrs=np.array([	["number",	"Rocks count",					int,	1,		1000],
											["z",		"Drop height (m)",				float,	2,		50	],
											["volume",	"Rocks volume (m<sup>3</sup>)",float,	0.1,	10	],
											["vz",		"Vertical velocity (m/s)",		float,	-10,	0	],
											["length",	"Length (m)",					float,	0.1,	3	],
											["width",	"Width (m)",					float,	0.1,	3	],
											["height",	"Height (m)",					float,	0.1,	3	]])
	def __init__(self, dt=0.02, engines=None,rocks_start_params_shapefile=None,**kwargs):
		super(Simulation, self).__init__(**kwargs)
		global current_simulation
		current_simulation=self
		self.forest_impact_model=None
		self.engines=engines or []
		self.use_siconos=False
		self.dt=dt
		self.iter=0
		self.rocks_start_params=[]
		self.current_start_params_id=0
		self.current_start_params_rock_id=0
		self.GUI_enabled=False
		if(rocks_start_params_shapefile is not None):
			self.set_rocks_start_params_from_shapefile(rocks_start_params_shapefile)
		self.automatic_engines_fill=None
		self.automatic_generate_forest={"enable":False, "density":0, "dhp_mean":30, "dhp_std":10}
		self.limit_fps=30 #used only with 3D view

	def set_rocks_start_params_from_shapefile(self,rocks_start_params_shapefile):
		try:
			self.rocks_start_params=[]
			shp_file=gdal.OpenEx(rocks_start_params_shapefile)
			layer0=shp_file.GetLayer(0)
			feature_nb=layer0.GetFeatureCount()
			for i in range(feature_nb): #loop on rocks start polygons
				self.rocks_start_params.append({}) #self.rocks_start_params is a list of dict
				feat=layer0.GetFeature(i).ExportToJson(as_object=True)
				shapely_poly = shapely.geometry.shape(feat['geometry'])
				shapely_poly = shapely.affinity.translate(shapely_poly,xoff=-self.terrain.Z_raster.xllcorner,yoff=-self.terrain.Z_raster.yllcorner)
				self.rocks_start_params[i]["shapely_polygon"]=shapely_poly
				self.rocks_start_params[i]["color"]=cp.dark2[i]
				rocks_properties = feat['properties']
				for k in list(rocks_properties.keys()):
					if(k not in Simulation.valid_input_shapefile_attrs[:,0]):
						rocks_properties.pop(k)
				self.rocks_start_params[i]["params"]=rocks_properties
		except Exception as e:
			Debug.error("The importation of rocks start params from shapefile failed:",str(e))
	
	def is_siconos_engine_capable(self):
		if not self.rocks_start_params:
			return False
		if not self.terrain:
			return False
		if not self.terrain.soil_params:
			return False
		for params in [rsp["params"] for rsp in self.rocks_start_params]:
			for p in ["height","length","width","z","vz","number"]:
				if not (p in params):
					return False
		for params in [sp["params"] for sp in self.terrain.soil_params]:
			for p in ["e","mu"]:
				if not p in params:
					return False
		return True
			
	def is_sphere_engine_capable(self):
		if not self.rocks_start_params:
			return False
		if not self.terrain:
			return False
		if not self.terrain.soil_params:
			return False
		for params in [rsp["params"] for rsp in self.rocks_start_params]:
			for p in ["volume","z","vz","number"]:
				if not (p in params):
					return False
		for params in [sp["params"] for sp in self.terrain.soil_params]:
			for p in ["roughness","bounce_model_number","R_t"]:
				if not p in params:
					return False
			if params["bounce_model_number"]==0 or params["bounce_model_number"]==1: #Classical and Pfeiffer
				if not "R_n" in params:
					return False
			elif(params["bounce_model_number"]==2):#Bourrier
				if not ("phi" in params and "v_half" in params):
					return False
		return True
	
	def stop_condition(self):
		return (len(self.output.contacts[-1])>1 and self.current_rock.vel.norm()<0.5)
	
	def save_to_file(self):
		"""
		Store (pickle to json) the simulation into a file whose path and name is the return result of :meth:`get_dir`. Note that the file will exclude all data that arises from calculations, this method is meant to save the simulation setup only.
		"""
		#store to local variables all data that shouldn't be saved to the file, then clear the corresponding simulation attributes :
		output=self.output
		self.output=None
		forest_impact_model=self.forest_impact_model
		self.forest_impact_model=None
		current_rock=self.current_rock
		self.current_rock=None
		if(self.terrain):
			terrain=copy.copy(self.terrain)
			self.terrain.faces=[]
			self.terrain.points=[]
			self.terrain.trees=[]
			self.terrain.faces_xyz_bb=None
			self.terrain.Z_raster.data={"Z":self.terrain.Z_raster.data["Z"]}	#only keep the Z data
		super(Simulation,self).save_to_file()	#actual file write is here
		#restore the not-saved data:
		self.output=output
		self.forest_impact_model=forest_impact_model
		if(self.terrain):
			self.terrain=terrain
		self.current_rock=current_rock
		#self.checkpoints=checkpoints
		Debug.Print("... DONE")
	
	def results_to_zipfile(self,filename=None):
		"""
		Create a zip file into the folder returned by :meth:`get_dir` named results.zip containing two text files. "stops.csv" contains info about rocks end position, and "checkpoints.csv" contains the checkpoints infos.
		"""
		if(filename is None):
			filename=self.get_dir()+"results.zip"
		zf = zipfile.ZipFile(filename, "w")
		pp=Postprocessings.Postprocessing(self)
		pp.run()
		outputs=pp.raster.output_to_asc(output_to_string=True)
		for field,output in outputs.items():
			zf.writestr(field+".asc",output)

		if(os.path.isfile(self.get_dir()+"terrain_overview.pdf")):
			zf.write(self.get_dir()+"terrain_overview.pdf",arcname="trajectories_overview.pdf")
		zf.close()
	
	def before_run_tasks(self):
		#set the simulation total number of rocks:
		self.nb_rocks=0
		for rsp in self.rocks_start_params:
			self.nb_rocks+=rsp['params']["number"]
		#call the parent function, which needs nb_rocks to be initialized:
		super(Simulation, self).before_run_tasks()
		
		self.current_start_params_rock_id=0
		self.current_start_params_id=0
		self.current_rock_id=0
		
		if(platrock.web_ui):
			self.terrain.populate_from_Z_raster()
			self.terrain.precompute_datas()
			self.automatic_generate_forest["enable"]=True
		self.terrain.set_faces_params_from_shapefile()
		if(self.automatic_engines_fill=="builtin"):
			self.engines=[	Engines.Verlet_update(use_cython=1,dist_factor=5),
							Engines.Contacts_detector(use_cython=1),
							Engines.Rock_terrain_nscd_basic_contact(),
							Engines.Rock_tree_nscd_basic_contact(),
							Engines.Nscd_integrator(use_cython=1)]
		elif(self.automatic_engines_fill=="siconos"):
			self.engines=[	Engines.Siconos(terrain=self.terrain, dt=0.001)]
		
		if(self.automatic_generate_forest["enable"]):
			self.terrain.generate_random_forest(self.automatic_generate_forest["density"],self.automatic_generate_forest["dhp_mean"],self.automatic_generate_forest["dhp_std"])
			self.enable_forest=len(self.terrain.trees)>0
		if self.enable_forest:
			self.terrain.trees_as_array=np.asarray([np.append(t.pos,t.dhp) for t in self.terrain.trees])
		#if forest, setup the impact model for 3D:
		if(self.enable_forest and self.forest_impact_model==None):
			self.forest_impact_model=BounceModels.Toe_Tree()

		#Define whether we use Siconos or not:
		if("Siconos" == self.engines[0].__class__.__name__):
			self.use_siconos=True
		else:
			self.use_siconos=False
		
		
		
	def before_rock_launch_tasks(self):
		super(Simulation, self).before_rock_launch_tasks()
		self.iter=0
		#select the right parameter set from self.rocks_start_params by using the sub-counter self.current_start_params_rock_id
		self.current_start_params_rock_id+=1
		if(self.current_start_params_rock_id>self.rocks_start_params[self.current_start_params_id]['params']["number"]):
			self.current_start_params_id+=1
			self.current_start_params_rock_id=1
		params=self.rocks_start_params[self.current_start_params_id]['params']
		polygon=self.rocks_start_params[self.current_start_params_id]['shapely_polygon']
		if(isinstance(polygon,shapely.geometry.MultiPolygon)):
			rand=self.random_generator.rand()
			ID=np.where(self.rocks_start_params[self.current_start_params_id]["multipoly_normalized_area_cumsum"]>rand)[0][0]
			polygon=polygon[ID]

		#find a rock start x,y pos in the polygon:
		min_x, min_y, max_x, max_y = polygon.bounds
		for i in range(1001):
			random_point = shapely.geometry.Point([self.random_generator.uniform(min_x, max_x), self.random_generator.uniform(min_y, max_y)])
			if(polygon.contains(random_point)):
				pos=Math.Vector3([random_point.x,random_point.y,0.])
				break
			if(i==1000):Debug.error("Unable to generate a point in the given polygon after 1000 tries.")
		#create the rock at [x,y,0]
		if(self.use_siconos):
			l=params.get("length",1)
			h=params.get("height",1)
			w=params.get("width",1)
			self.current_rock=Objects.Parallelepiped(lengths=[l,h,w],pos=pos)
			self.current_rock.ori=quaternion.from_euler_angles(self.random_generator.rand(3)*2*np.pi)
		else:
			self.current_rock=Objects.Sphere(volume=params["volume"],pos=pos)
		#find the face below the rock then set the rock altitude:
			#1- use the faces bounding boxes in the (x,y) plane to get a rough list
		mins_check_outside = (pos[0:2]<self.terrain.faces_xyz_bb[:,0:3:2]).any(axis=1)
		maxs_check_outside = (pos[0:2]>self.terrain.faces_xyz_bb[:,1:5:2]).any(axis=1)
		inside_mask = np.logical_not(np.logical_or(mins_check_outside, maxs_check_outside))
			#2- loop on the faces list to get the face right below the rock
		for f in self.terrain.faces[inside_mask]:
			#find the corresponding face:
			if(f.is_point_inside_2D(self.current_rock)):	#find the Z position of the rock right above the corresponding face
				self.current_rock.pos[2] = -(f.normal[0]*(random_point.x-f.points[0].pos[0]) + f.normal[1]*(random_point.y-f.points[0].pos[1]))/f.normal[2]+f.points[0].pos[2]+params["z"]+self.current_rock.radius
				break
		if(self.current_rock.pos[2]==0): Debug.error("Couldn't find a suitable height for the current rock.")
		
		self.current_rock.vel=Math.Vector3([self.random_generator.rand()*0.01,self.random_generator.rand()*0.01,params["vz"]])
		if(self.use_siconos):
			self.engines[0].add_rock(self.current_rock)
		self.output.add_rock(self.current_rock)
		self.output.add_contact(self.current_rock,Math.Vector3([0.,1.,0.]),Out.START) #store the rock initial position
		if(self.GUI_enabled):
			self.GUI.draw_rock(self.current_rock)
		# Reset forest:
		if(self.terrain):
			for tree in self.terrain.trees:
				tree.active=True
				tree.color=[0,0.8,0]
	
	def after_rock_propagation_tasks(self):
		type(self).__bases__[0].after_rock_propagation_tasks(self)
		if(self.current_rock.out_of_bounds):
			self.output.add_contact(self.current_rock,Math.Vector3([0.,0.,0.]),Out.OUT)
		else:
			self.output.add_contact(self.current_rock,Math.Vector3([0.,0.,0.]),Out.STOP)
	
	def after_successful_run_tasks(self):
		super(Simulation, self).after_successful_run_tasks()
		if platrock.web_ui:
			import platrock.GUI.Plot3D as Plot3D
			Plot3D.clear()
			Plot3D.draw_terrain(self,with_polygons=True)
			Plot3D.draw_sample_trajectories(self,50)
			Plot3D.fig.savefig(self.get_dir()+'terrain_overview.svg',bbox_inches='tight')
			Plot3D.fig.savefig(self.get_dir()+'terrain_overview.pdf',bbox_inches='tight')
			Plot3D.plt.close()
			self.results_to_zipfile()
	
	def after_all_tasks(self):
		super(Simulation, self).after_all_tasks()

	def rock_propagation_tasks(self):
		#Don't call the parent function here as it would slow down a bit. It would be useless as there is nothing in the parent class (nothing common between TwoD and ThreeD)
		if(not self.status=="running"):
			if(self.GUI_enabled):
				self.GUI.V.window.start_stop_button.setText("RESUME")
			time.sleep(0.1)
			return
		if(self.GUI_enabled):
			for f in self.current_rock.verlet_faces_list:
				f.color=[0.68235294, 0.5372549 , 0.39215686]
		for E in self.engines:
			if( (not E.dead) and self.iter%E.iter_every==0):E.run(self)
		if(self.GUI_enabled):
			for f in self.current_rock.verlet_faces_list:
				f.color=[0.82352941, 0.41176471, 0.11764706]
			if(self.current_rock.terrain_active_contact):
				self.current_rock.terrain_active_contact.face.color=np.random.rand(3)
			if(time.time()-self.GUI.V.time_last_updated>1/self.GUI.V.fps):
				self.GUI.V.time_last_updated=time.time()
				self.GUI.updateTO(self.current_rock)
				self.GUI.updateTerrainColors(self.terrain)
				self.GUI.V.camera.SetFocalPoint(self.current_rock.pos)
				self.GUI.V.window.start_stop_button.setText("PAUSE")
				self.GUI.V.iren.Render()
				time.sleep(0.01)
		if(self.stop_condition()):
			self.current_rock.is_stopped=True
			if(self.use_siconos):
				self.engines[0].remove_rock()
			Debug.info("Rock stopped")
			if(self.GUI_enabled and (not self.use_siconos)):
				self.GUI.V.ren.RemoveActor(self.current_rock.bounding_box.VTK_actor)
				for f in self.current_rock.verlet_faces_list:
					f.color=[0.68235294, 0.5372549 , 0.39215686]
		self.iter+=1



















"""
Copyright (c) 2017, 2018, 2019 Irstea
Copyright (c) 2017, 2018, 2019 François Kneib
Copyright (c) 2017, 2018, 2019 Franck Bourrier
Copyright (c) 2017, 2018, 2019 David Toe
Copyright (c) 2017, 2018, 2019 Frédéric Berger
Copyright (c) 2017, 2018, 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""


#TODO :
# -globally remove all Debug.Print in PlatRock, repace by one of the three log levels.

import multiprocessing,logging,sys,os,subprocess

#Message importance levels :
INFO=logging.INFO #20
WARNING=logging.WARNING #30
ERROR=logging.ERROR #40

#Actual log level : (log messages whose importance > current_level)
current_level=0


# Get the logger corresponding to the current process PID,
# Or create one if it doesn't exist (this is the behavior of "logging.getLogger")
def get_logger():
	return logging.getLogger(str(multiprocessing.current_process().pid))	#create a logger for this thread
	
# Get or create a logger based on the process PID
# Clear its handlers
# Add a new handler
def add_logger(filename=None):
	logger = get_logger()	# get or create the logger
	logger.handlers=[]		# remove all the handlers
	if(filename==None or multiprocessing.current_process().name=='MainProcess'): handler = logging.StreamHandler()	#if the process is main (no multithreading, so no WebUI), log to console
	else: 
		if(os.path.exists(filename)): subprocess.call(["rm",filename])
		handler = logging.FileHandler(filename)									#if the process is a child, log to file
	logger.addHandler(handler)	#add the handler to the logger
	logger.propagate = False #avoid duplicate logging to the console
	logger.setLevel(current_level)
	logger.name="PlatRock-WebUI PID="+multiprocessing.current_process().name

def args_to_str(*args):
	S=""
	for arg in args:
		S+=str(arg)+" "
	return S

# Concatenate the input args, get the logger then log:
def info(*args):
	get_logger().info(args_to_str(*args))

Print=info #for retro-compatibility of the old Debug version.
	
def warning(*args):
	get_logger().warning(args_to_str(*args))

def error(*args):
	get_logger().error(args_to_str(*args))

def do_nothing(*args):
	pass

# Override the previous functions with "do_nothing" depending on the log level :
def init(level):
	global current_level
	current_level=level
	if(level>INFO):
		global info,Print
		info=do_nothing
		Print=do_nothing
	if(level>WARNING):
		global warning
		warning=do_nothing
	if(level>ERROR):
		global error
		error=do_nothing



"""
Copyright (c) 2019 François Kneib
Copyright (c) 2019 Franck Bourrier
Copyright (c) 2019 David Toe
Copyright (c) 2019 Frédéric Berger
Copyright (c) 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""

import platrock.Common.Simulations
import platrock.TwoD.Simulations
from   platrock.Common.PyUtils import FriendImporter
import platrock.Common.Math as Math
import platrock.Common.Outputs as Out
from . import Objects
import numpy as np
import copy
from siconos.mechanics.collision.bullet import SiconosBulletCollisionManager, SiconosBulletOptions
from siconos.kernel import DynamicalSystem, NonSmoothDynamicalSystem, MoreauJeanOSI, TimeDiscretisation, RollingFrictionContact, TimeStepping, SICONOS_TS_NONLINEAR, NewtonImpactRollingFrictionNSL
from siconos.mechanics.collision import RigidBody2dDS, SiconosContactorSet, SiconosContactor, SiconosConvexHull2d
from siconos.io.io_base import MechanicsIO

from siconos.io.mechanics_run import MechanicsHdf5Runner
from siconos.mechanics.collision.tools import Contactor

#FIXME: remove these:
from siconos.kernel import FrictionContact, NewtonImpactFrictionNSL

class Simulation(platrock.Common.Simulations.Simulation,FriendImporter):
	webui_typename="PlatRock 2D-Shape"
	valid_rocks_start_params=copy.deepcopy(platrock.TwoD.Simulations.Simulation.valid_rocks_start_params)
	#We imported valid_rocks_start_params from TwoD.Simulations.Simulation, but don't want rocks_volume to be a valid input parameter:
	ID=np.where(valid_rocks_start_params[:,0]=="rocks_volume")[0][0]
	valid_rocks_start_params=np.concatenate((valid_rocks_start_params[:ID],valid_rocks_start_params[ID+1:]))
	#And we want to add these params:
	valid_rocks_start_params=np.concatenate((valid_rocks_start_params,
		np.array([
			["random_rocks_ori",	"Random rocks orientation",				bool,		False,		True	],
			["rocks_ori",			"Rocks orientation (degrees)",			int,		0,			359		]
		])
	))
	
	friendClass=platrock.TwoD.Simulations.Simulation
	friendImportNames=[
		"nb_rocks",
		"rocks_density",
		"rocks_x",
		"rocks_z",
		"rocks_vx",
		"rocks_vz",
		"rocks_angVel",
		"update_checkpoints",
		"checkpoints",
		"save_to_file",
		"after_rock_propagation_tasks",
		"after_successful_run_tasks",
		"results_to_zipfile",
		"get_stops_cdf"
	]

	def __init__(self,*args,rocks_start_params={},checkpoints_x=[],**kwargs):
		super(Simulation,self).__init__(*args,**kwargs) #first parent
		FriendImporter.__init__(self, rocks_start_params=rocks_start_params,checkpoints_x=checkpoints_x)
		
		self.random_rocks_ori=False
		self.rocks_ori=0.
		
		#OVERRIDE THE DEFAULT PARAMS, IF AVAILABLE IN rocks_start_params dict.
		for k in rocks_start_params.keys():
			if(k in self.valid_rocks_start_params[:,0]):
				self.__dict__[k]=rocks_start_params[k]
		
		self.rocks_shape_type=Objects.Ellipse #mainly for webui
		self.rocks_shape_params={}
		self.export_hdf5=False
		self.last_record_pos=Math.Vector2([np.inf,np.inf])
	
	def terrain_to_siconos_hdf5(self):
		with MechanicsHdf5Runner(self.name+".hdf5") as io:
			for seg in self.terrain.segments:
				io.add_convex_shape('Seg'+str(seg.index),seg.siconos_points)
				io.add_object('seg'+str(seg.index), [Contactor('Seg'+str(seg.index))] ,translation=[0,0])
	
	def rock_to_siconos_hdf5(self,r):
		with MechanicsHdf5Runner(self.name+".hdf5",mode="r+") as io:
			io._number_of_dynamic_objects=self.current_rock_number #tweak this as add_object() uses it to index objects in hdf5.
			io.add_convex_shape('Rock'+str(self.current_rock_number),r.vertices)
			io.add_object('rock'+str(self.current_rock_number), [Contactor('Rock'+str(self.current_rock_number))], mass=1., inertia=1., translation=r.pos.tolist())
			
	def setup_siconos(self,dt=0.02):
		self.io = MechanicsIO()
		bulletOptions = SiconosBulletOptions()
		bulletOptions.worldScale = 1.0
		bulletOptions.contactBreakingThreshold = 0.04
		bulletOptions.dimension = 1
		self.interman=SiconosBulletCollisionManager(bulletOptions)
		self.nsds = NonSmoothDynamicalSystem(0,np.inf)
		DynamicalSystem.resetCount(0)
		self.osi = MoreauJeanOSI(0.50001)
		self.osi.setConstraintActivationThreshold(0.0)
		self.timedisc=TimeDiscretisation(0, dt)#FIXME FIXME FIXME : set DT here !
		
		#self.osnspb=RollingFrictionContact(3)
		self.osnspb=FrictionContact(2)
		self.osnspb.setMaxSize(30000)
		self.osnspb.setMStorageType(1)
		
		solverOptions = self.osnspb.numericsSolverOptions()
		solverOptions.iparam[0] = 1000		#iter max tester 100
		solverOptions.dparam[0] = 1e-5		#tolerance tester 1e-3
		self.osnspb.setNumericsVerboseMode(False)
		self.osnspb.setKeepLambdaAndYState(True)
		
		self.sim=TimeStepping(self.nsds,self.timedisc)
		self.sim.insertIntegrator(self.osi)
		self.sim.insertNonSmoothProblem(self.osnspb)
		self.sim.insertInteractionManager(self.interman)
		self.sim.setNewtonOptions(SICONOS_TS_NONLINEAR)
		self.sim.setNewtonMaxIteration(1)
		self.sim.setNewtonTolerance(1e-5) #tester 1e-3
		self.sim.setDisplayNewtonConvergence(False)
		
	def add_terrain(self):
		terrain_points=self.terrain.get_points()
		terrain_min_z=terrain_points[:,1].min()-10
		for seg in self.terrain.segments:
			seg.siconos_points=np.array([	[seg.points[0][0],terrain_min_z],
								[seg.points[1][0],terrain_min_z],
								[seg.points[1][0],seg.points[1][1]],
								[seg.points[0][0],seg.points[0][1]]
							])
			dims = [	seg.siconos_points[:,0].max() - seg.siconos_points[:,0].min(),
						seg.siconos_points[:,1].max() - seg.siconos_points[:,1].min()]
			shp=SiconosConvexHull2d(seg.siconos_points)
			shp.setInsideMargin(min(dims)*0.02)
			shp.setOutsideMargin(0)
			csetpos = [0,0,0,1,0,0,0]
			cset = SiconosContactorSet()
			cset.append(SiconosContactor(shp, csetpos, seg.index+1))
			self.interman.insertStaticContactorSet(cset,csetpos)
			#seg.nsl=NewtonImpactRollingFrictionNSL(seg.e, 0., seg.mu, seg.mu_r, 3) # en, et, mu, mu_r, ndim
			seg.nsl=NewtonImpactFrictionNSL(seg.e, 0., seg.mu, 2) # en, et, mu, ndim
			self.interman.insertNonSmoothLaw(seg.nsl, seg.index+1, 0)
		
	
	def add_rock(self):
		self.reset_rock_kinematics()
		r=self.current_rock
		self.DS=RigidBody2dDS(r.pos.tolist()+[r.ori],r.vel.tolist()+r.angVel.tolist(), r.mass, r.I)
		self.DS.setNumber(self.current_rock_number+1)
		self.DS.setUseContactorInertia(False)
		self.DS.setAllowSelfCollide(False)
		cset = SiconosContactorSet()
		cset.append(SiconosContactor(r.get_siconos_shape(), [0,0,0,1,0,0,0], 0))
		self.DS.setContactors(cset)
		self.nsds.insertDynamicalSystem(self.DS)
		self.nsds.setName(self.DS, 'rock'+str(self.current_rock_number))
		self.DS.setFExtPtr([0, - self.DS.scalarMass() * self.gravity,0])
		if self.export_hdf5 :
			self.rock_to_siconos_hdf5(self.current_rock)
	
	def reset_rock_kinematics(self):
		r=self.current_rock
		r.setup_kinematics(x=self.rocks_x, height=self.rocks_z, vel=Math.Vector2([self.rocks_vx,self.rocks_vz]), angVel=self.rocks_angVel)
		if(self.random_rocks_ori):
			r.ori=self.random_generator.rand()*np.pi*2
		else:
			r.ori=np.radians(self.rocks_ori)
		r.update_current_segment(self.terrain)
		r.pos[1]+=r.current_segment.get_y(r.pos[0])
	
	def remove_rock(self):
		self.nsds.removeDynamicalSystem(self.DS)
		self.interman.removeBody(self.DS)
		

	def before_run_tasks(self):
		super(Simulation, self).before_run_tasks()
		self.setup_siconos()
		self.add_terrain()
		if self.export_hdf5 :
			self.terrain_to_siconos_hdf5()
			#if self.export_hdf5 :
			self.hdf5_io = MechanicsHdf5Runner(self.name+".hdf5",mode='r+')
			self.hdf5_io.__enter__()
			self.hdf5_io._nsds=self.nsds
			self.hdf5_io._simulation=self.sim

	def before_rock_launch_tasks(self):
		import time
		super(Simulation, self).before_rock_launch_tasks()
		self.add_rock()
		self.reset_rock_kinematics()
		self.last_record_pos=Math.Vector2([np.inf,np.inf])
		self.output.add_rock(self.current_rock)
		self.output.add_contact(self.current_rock,Math.Vector2([0.,0.]),Out.START) #store the rock initial position
		

	def rock_propagation_tasks(self):
		self.sim.computeOneStep()
		if self.export_hdf5 :
			self.hdf5_io.output_dynamic_objects()
			self.hdf5_io.output_velocities()
			self.hdf5_io.output_contact_forces()
			self.hdf5_io.output_solver_infos()
			self.hdf5_io._out.flush()
		pos=self.DS.q()
		vel=self.DS.velocity()
		
		#Check out_of_bounds
		if	pos[0]<self.terrain.segments[0].points[0][0] or \
			pos[0]>self.terrain.segments[-1].points[1][0] or \
			pos[1]<self.terrain.get_y_range()[0] :
			self.current_rock.out_of_bounds=True
			self.current_rock.is_stopped=True
		
		#Record trajectory
		if (self.last_record_pos-pos[0:2]).norm() > max(self.current_rock.dims)*5 :
			self.current_rock.pos=Math.Vector2(pos[0:2])
			self.current_rock.ori=pos[2]
			self.current_rock.vel=Math.Vector2(vel[0:2])
			self.current_rock.angVel=Math.Vector1([vel[2]])
			self.output.add_contact(self.current_rock,Math.Vector2([0,0]),Out.MOTION)
			if self.current_rock.vel.norm() < 1e-2:
				self.current_rock.is_stopped=True
				self.remove_rock()
		self.sim.clearNSDSChangeLog()
		self.sim.nextStep()

	def after_all_tasks(self):
		super(Simulation, self).after_all_tasks()
		if self.export_hdf5 :
			self.hdf5_io.__exit__(None, None, None)



















"""
Copyright (c) 2017, 2018, 2019 Irstea
Copyright (c) 2017, 2018, 2019 François Kneib
Copyright (c) 2017, 2018, 2019 Franck Bourrier
Copyright (c) 2017, 2018, 2019 David Toe
Copyright (c) 2017, 2018, 2019 Frédéric Berger
Copyright (c) 2017, 2018, 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import copy


def from_terrain(terrain,cell_length,xllcorner=0,yllcorner=0):
	"""
	Create a new raster from the terrain only. /!\ if the terrain was generated from an asc, please use "from_asc" then "from_raster" to avoid shifting of the origin.
	"""
	r=Raster()
	xmin=terrain.points_as_array[:,0].min()
	xmax=terrain.points_as_array[:,0].max()
	ymin=terrain.points_as_array[:,1].min()
	ymax=terrain.points_as_array[:,1].max()
	r.cell_length=float(cell_length)
	r.X=np.arange(xmin,xmax+cell_length,cell_length)
	r.Y=np.arange(ymin,ymax+cell_length,cell_length)
	r.nx=len(r.X)
	r.ny=len(r.Y)
	r.xllcorner=xllcorner
	r.yllcorner=yllcorner
	return r

def from_raster(raster,cell_length=None):
	"""
	Copy a raster without the data, optionnaly change the cell_length (=remap the raster)
	"""
	r=copy.deepcopy(raster)
	if(cell_length is not None):
		r.cell_length=cell_length
		r.X=np.arange(raster.X[0],raster.X[-1]+cell_length,cell_length)
		r.Y=np.arange(raster.Y[0],raster.Y[-1]+cell_length,cell_length)
		r.nx=len(r.X)
		r.ny=len(r.Y)
	r.data={}
	return r

def from_asc(filename,data_name,header_lines_count=6):
	r=Raster()
	#LOAD FILE AND CREATE RASTER ARRAYS :
	#load header:
	r.header_data={"cellsize":None,"NODATA_value":None,"xllcorner":None,"yllcorner":None,"ncols":None,"nrows":None}
	with open(filename,'r') as f:
		for i in range(header_lines_count):
			data=f.readline().split()
			if data[0] in r.header_data:
				try:
					value=int(data[1])
				except:
					value=float(data[1])
				r.header_data[data[0]]=value
	r.cell_length=r.header_data["cellsize"]
	r.xllcorner=r.header_data["xllcorner"]
	r.yllcorner=r.header_data["yllcorner"]
	#load terrain:
	data = np.loadtxt(filename, skiprows=header_lines_count).T
	data = np.flip(data,axis=1) #numpy reads the file from top to bottom, but the origin of the ascii raster is at the bottom (right)
	r.nx=np.shape(data)[0]
	r.ny=np.shape(data)[1]
	r.X=np.arange(0,r.nx)*r.cell_length
	r.Y=np.arange(0,r.ny)*r.cell_length
	#store in raster data:
	r.data[data_name]=data
	return r
	

class Raster(object):
	def __init__(self):
		self.X=None
		self.Y=None
		self.nx=None
		self.ny=None
		self.cell_length=None
		self.xllcorner=None
		self.yllcorner=None
		self.data={}
	
	def get_indices_from_coords(self,coords):
		#coords must be np.array([x,y])
		indices_coords=(coords/self.cell_length).astype(int)					#indices of the cell ([xi,yi])
		return indices_coords
	
	def add_data_grid(self,name,type,default_value=0):
		self.data[name]=np.empty([self.nx,self.ny],dtype=type)
		if(type==list):
			for i in range(np.shape(self.data[name])[0]):
				for j in range(np.shape(self.data[name])[1]):
					self.data[name][i,j]=[]
		else:
			self.data[name].fill(default_value)
		return self.data[name]
	
	def data_to_mean(self,data_name):
		rd=self.data[data_name]
		a=np.zeros(rd.shape)
		for i in range(rd.shape[0]):
			for j in range(rd.shape[1]):
				if(len(rd[i,j])):
					a[i,j]=sum(rd[i,j])/len(rd[i,j])
		return a
	
	def get_asc_header_string(self):
		values={"ncols":int(self.nx),
				"nrows":int(self.ny),
				"xllcorner":int(self.xllcorner),
				"yllcorner":int(self.yllcorner),
				"cellsize":self.cell_length,
				"NODATA_value":self.header_data["NODATA_value"]
				}
		S=""
		for k in values:
			S+=str(k).ljust(14)+str(values[k])+"\n"
		return S
	
	def output_to_asc(self,output_to_string=False,fields=None):
		if not fields:
			fields=list(self.data.keys())
			fields.remove("stops")
			fields.remove("Z")
		arrays={}
		for fi in fields:
			arrays[fi]=np.empty([self.nx,self.ny],dtype=float)
			arrays[fi].fill(self.header_data["NODATA_value"])
		for x in range(self.nx):
			for y in range(self.ny):
				if(self.data["crossings"][x,y]==0):
					continue
				else:
					for fi in fields:
						arrays[fi][x,y]=np.asarray(self.data[fi][x,y]).mean()
		class output_buff():
			def __init__(self):
				self.buff=""
			def write(self,what):
				self.buff+=what
			def __repr__(self):
				return self.buff

		outputs={}
		for fi in fields:
			if(output_to_string):
				outputs[fi]=output_buff()
			else:
				outputs[fi]=open(str(fi)+"_mean.asc",'w')
			outputs[fi].write(self.get_asc_header_string())
			np.savetxt(outputs[fi], np.flip(arrays[fi],axis=1).T, fmt="%.2f")	#flip and transpose to translate from numpy representation to ascii grid representation
			if(output_to_string):
				outputs[fi]=outputs[fi].buff
			else:
				outputs[fi].close()
		if(output_to_string):
			return outputs


















"""
Copyright (c) 2017, 2018, 2019 Irstea
Copyright (c) 2017, 2018, 2019 François Kneib
Copyright (c) 2017, 2018, 2019 Franck Bourrier
Copyright (c) 2017, 2018, 2019 David Toe
Copyright (c) 2017, 2018, 2019 Frédéric Berger
Copyright (c) 2017, 2018, 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""

"""
This module is used by the TwoD model. It handles all the Objects types
"""
#FIXME : rewrite the override mechanism and allow to set default segments parameters from the launch script (as in 3D).

import numpy as np
import platrock.Common.Debug as Debug
import copy,sys
from . import Geoms
import platrock.Common.Math as Math
import platrock.Common.ColorPalettes as cp
import platrock.Common.Outputs as Out

class Rock(object):
	"""A falling rock.
	
	Args:
		pos (:class:`Common.Math.Vector2` [float,float]): position along x,z
		vel (:class:`Common.Math.Vector2` [float,float]): velocity along x,z
		angVel (float): angular velocity
		volume (float): volume
		density (float): density
		I (float): inertia
		radius (float): radius
		mass (float): mass
		is_stopped (boolean): flag triggered when the stopping condition is reached
		current_segment (:class:`Segment`): the current segment that is vertically under the rock
		flying_direction (int): -1 if the rock is moving towards -x, +1 if the rock is moving towards +x
		color (list [float,float,float]): the rock RGB color, each compound being between 0. and 1.
		contacts (list [:class:`Contact` , ...]): list of all contacts that the rock has had with soil or trees.
			A first contact is added at the initial position of the rock, even though it's not a real contact
		out_of_bounds (bool): set to true during the simulation if the rock went out of the terrain
	"""
	def __init__(self,x=0.,height=1.,vel=[1.,0.],angVel=0.,volume=1.,density=2800.,I=None):
		"""
		Constructor
		
		Args:
			x (float): initial position along the x axis. Note that the coordinates system used is the one after the terrain is eventually cleaned, shifted and reversed.
			height (float): initial height, relative to 0
		"""
		
		self.setup_kinematics(x,height,vel,angVel)
		self.volume=volume
		self.density=density
		
		self.radius = (3.*self.volume/np.pi/4.)**(1./3.)
		self.mass = self.volume*self.density
		self.I = I or 2./5.*self.mass*self.radius**2
		self.A = self.mass/(self.mass+self.I/self.radius**2) # see Azzoni et al. 1995
		self.v_square=0 #needed by the rolling aglorithm
	
	def setup_kinematics(self,x=None,height=None,vel=None,angVel=None):
		self.pos=Math.Vector2([x,height])
		self.vel=Math.Vector2(vel)
		self.angVel=Math.Vector1(angVel)
		self.is_stopped=False
		self.current_segment=None
		self.flying_direction=0
		self.color=[np.random.rand(),np.random.rand(),np.random.rand()]
		self.out_of_bounds=False
	
	def update_current_segment(self,input):
		"""
		Update the :attr:`current_segment` of this rock.
		
		Args:
			input (:class:`Segment` | :class:`Terrain`): the new segment or the terrain from which to compute.
		"""
		if(isinstance(input,Segment)):
			self.current_segment=input
			Debug.Print("CURRENT SEGMENT IS SET TO N",self.current_segment.index)
		elif(isinstance(input,Terrain)):
			terrain_points=input.get_points()
			seg=input.segments[np.where(terrain_points[:,0]<=self.pos[0])[0][-1]]
			if abs(self.pos[0]-seg.points[0][0]) < 1e-10: #we are exactly at a point, use velocity to choose between segment before or after.
				self.current_segment = seg if self.flying_direction>0 else input.segments[seg.index-1]
			else :
				self.current_segment=seg
			Debug.Print("CURRENT SEGMENT IS SET TO N",self.current_segment.index)
		else:
			Debug.Print("ERROR, update_current_segment_number called with a wrong input parameter.")
	def update_flying_direction(self):
		"""
		Update the :attr:`flying_direction` from the velocity.
		"""
		self.flying_direction=int(np.sign(self.vel[0]))
		Debug.Print("Rock direction is set to",self.flying_direction)
	
	def move(self,arrival_point,s,segment):
		#update position :
		self.pos=arrival_point
		self.update_current_segment(segment)
		if(self.vel.norm()<0.5 and s.output.contacts[-1].shape[0]>1):
			Debug.Print("ROCK STOPPED")
			self.is_stopped=True
	
	def fly(self,arrival_point,s,segment):
		"""
		Update the position and the velocity of the rock. Also contains the rock stop condition, so sets the :attr:`is_stopped` flag if necessary.
		
		Args:
			arrival_point (:class:`Common.Math.Vector2` [float,float]): point to reach after fly
			gravity (float): gravity to apply during flight #FIXME: DOC
			segment (:class:`Segment`): arrival segment
		"""
		#update velocity regarding the  arrival point :
		self.vel[1]=-s.gravity*(arrival_point[0]-self.pos[0])/self.vel[0] + self.vel[1]
		Debug.Print("Fly to",arrival_point,",new vel is",self.vel)
		self.move(arrival_point,s,segment)
	
	def roll(self,s,azzoni_roll,arrival_point):
		Debug.info("Roll to ",arrival_point)
		if(azzoni_roll.until_stop):
			self.vel*=0
			self.angVel*=0
			self.move(arrival_point,s,self.current_segment)
		else:
			self.vel=azzoni_roll.get_vel(arrival_point)
			self.angVel = Math.Vector1(- self.flying_direction * self.vel.norm()/self.radius)
			if self.flying_direction>0 and arrival_point[0]>=self.current_segment.points[1][0] :
				next_seg_id=self.current_segment.index+1
			elif self.flying_direction<0 and arrival_point[0]<=self.current_segment.points[0][0] :
				next_seg_id=self.current_segment.index-1
			else: next_seg_id=self.current_segment.index
			if next_seg_id<0 or next_seg_id>len(s.terrain.segments)-1 :
				self.out_of_bounds=True
				self.is_stopped=True
				next_segment = self.current_segment
			else:
				next_segment = s.terrain.segments[next_seg_id] 
				
			self.move(arrival_point,s,next_segment)
	
	def bounce(self,s,segment,disable_roughness=False):
		if(s.override_rebound_params):
			bounce_model_number=s.override_bounce_model_number
		else:
			bounce_model_number=segment.bounce_model_number
		bounce_model=s.number_to_model_correspondance[bounce_model_number]
		bounce_model.run(self,segment,disable_roughness)
		return bounce_model.updated_normal

class Contact(object):
	"""
	A contact, used to keep history and store data of a single contact.
	
	Args:
		pos (:class:`Common.Math.Vector2` [float,float]): position along x,z
		vel (:class:`Common.Math.Vector2` [float,float]): velocity along x,z
		normal (:class:`Common.Math.Vector2` [float,float]): normal of the segment facing upwards seen by the rock after the roughness was applied
	"""
	def __init__(self,r,normal,type=None):
		self.rock_pos=r.pos.copy()
		self.rock_output_vel=r.vel.copy()
		self.normal=normal.copy()
		self.rock_output_angVel=r.angVel.copy()
		self.type=type

class Segment(object):
	"""
	A segment of the terrain.
		
	Args:
		points (:class:list [[x0, z0], [x1, z1]]): start and end points	
		bounce_model_number (int): integer representing the bounce model	
		R_t (float): tangential restitution coefficient [0.-1.]
		R_n (float): normal restitution coefficient [0.-1.]
		roughness (float): roughness #FIXME: definition?
		v_half (float): v_half #FIXME: definition?
		phi (float): phi #FIXME: definition?
		trees_density (float): the 3D density of the trees, in trees/hectare #FIXME : check this
		dhp_mean (float): mean dhp, diameter at breast height
		dhp_std (float): dhp standard deviation, diameter at breast height
		slope_gradient (float): the slope gradient, equals to the :math:`a` coefficient of the line
		index (int): index of the segment (they are continuously indexed through the terrain)
	"""
	def __init__(self, start_point, end_point, bounce_model_number=None, R_t=None, R_n=None, roughness=None,mu_r=None, v_half=None, phi=None, trees_density=None, dhp_mean=None, dhp_std=None):
		"""
		Constructor
		
		Args:
			start_point (:class:`Common.Math.Vector2` [x,z]): start point of the segment
			end_point (:class:`Common.Math.Vector2` [x,z]): end point of the segment
		"""
		self.points=np.array([start_point,end_point])
		self.bounce_model_number=bounce_model_number
		self.R_t=R_t
		self.R_n=R_n
		self.roughness=roughness
		self.mu_r=mu_r
		self.v_half=v_half
		self.phi=phi
		self.trees_density=trees_density
		self.dhp_mean=dhp_mean
		self.dhp_std=dhp_std
		self.index=-1
		self.set_geometrical_quantities()
	
	def set_geometrical_quantities(self):
		self.branch=Math.Vector2(self.points[1]-self.points[0])
		if(abs(self.branch[0])<1e-5):
			self.slope_gradient=0.
		else:
			self.slope_gradient=self.branch[1]/self.branch[0]
		self.normal=self.branch.rotated(np.pi/2.).normalized()
		self.slope=np.arctan(self.slope_gradient)
	
	def get_y(self,x):
		"""
		Get the y value from a x value. Warning, x must be between segment start and end.
		
		"""
		return self.points[0][1]+self.slope_gradient*(x-self.points[0][0])
			

class Terrain(object):
	"""
	A 2D terrain made of segments
	
	Args:
		segments (list [:class:`Segment`, ...]): the successive segments forming the terrain
		rebound_models_available (list [int,int,int]): A list of available rebound models regarding the per-segment input parameters given. Modified by check_segments_parameters_consistency() method.
		forest_available (bool): whether the forest is available in the terrain. Modified by check_segments_parameters_consistency() method.
	"""
	valid_input_attrs=np.array([	["R_n",					"R<sub>n</sub>",				float,	0,		0.95	],
									["R_t",					"R<sub>t</sub>",				float,	0,		0.95	],
									["roughness",			"Roughness",					float,	0,		1.		],
									["bounce_model_number",	"Rebound model",				int,	0,		2		],
									["mu_r",				"μ<sub>r</sub>",				float,	0,		100		],
									["phi",					"φ",							float,	0,		40		],
									["v_half",				"V<sub>1/2</sub>",				float,	0,		50		],
									["trees_density",		"Trees density  (trees/ha)",	float,	0,		5000	],
									["dhp_mean",			"Mean trees ⌀ (cm)",			float,	5,		200		],
									["dhp_std",				"σ(trees ⌀)  (cm)",				float,	0.01,	1000	]])
	def __init__(self,file=None,segments_type=Segment):
		self.segments=[]
		self.params_zones=[]
		self.rebound_models_available=[]#A list of available rebound models regarding the per-segment input parameters given. Modified by check_segments_parameters_consistency() method.
		self.forest_available=False
		if(isinstance(file,str)):
			self.import_from_csv(file,segments_type)
			self.check_segments_continuity()
			self.check_segments_parameters_consistency()
			self.index_segments()
			self.set_params_zones()
	
	def import_from_csv(self,file,segments_type):
		"""
		Constructor
		
		Args:
			file (string): path to the terrain file. This file is a basic csv text file, which contains points (X,Z) and per-segment parameters. It is formed as follows:
			
				+-+-+-------------------+---+---+---------+------+---+-------------+--------+-------+
				|X|Z|bounce_model_number|R_t|R_n|roughness|v_half|phi|trees_density|dhp_mean|dhp_std|
				+-+-+-------------------+---+---+---------+------+---+-------------+--------+-------+

				The first line of the file MUST be the columns names, which are CASE-SENSITIVE but the sequence can be swapped. The data must be rectangular, meaning that the table must be fullfilled without "no-data" possibility. You can import a terrain with only X and Z columns, but in this case you will have to complete the segments parameters in your script before launching the simulation.
		
		"""
		### all these code lines are needed to clean the terrain.
		input_array=np.genfromtxt(file,names=True, dtype=float)
		if(input_array['Z'][0]-input_array['Z'][-1]<0): # mirror terrain through X if the terrain is ascendent
			input_array['X']=-input_array['X']
			input_array['X']-=input_array['X'].min()
		if(input_array['X'][-1]-input_array['X'][0]<0):	# reverse the values through X if they are inversed
			input_array=np.flip(input_array,axis=0)
		# add horizontal segments at the right and left
		#input_array=np.append(input_array,[input_array[-1]]) ; input_array[-1]["X"]+=0.1*(input_array[-1]["X"]-input_array[0]["X"])
		#input_array=np.insert(input_array,0,[input_array[0]]) ; input_array[0]["X"]-=0.1*(input_array[-1]["X"]-input_array[0]["X"])
		
		### BEGIN REMOVE CAVITIES
		# 1:sort indices by X position in a new list and delete indices that creates cavities
		cleaned_indices=[]
		sorted_indices=np.argsort(input_array['X'])
		unsorted_indices=list(range(len(sorted_indices)))
		counter=0
		while counter < len(sorted_indices):
			i=sorted_indices[counter]
			if(i==unsorted_indices[0]):		# no cavity
				cleaned_indices.append(i)
				unsorted_indices.pop(0)
			else:							# there is a cavity here. Compute V1 and V2 to know whether its a cavity towards left or right of the terrain
				V1=Math.Vector2([input_array[i]['X'],input_array[i]['Z']])-Math.Vector2([input_array[unsorted_indices[0]]['X'],input_array[unsorted_indices[0]]['Z']])
				V2=Math.Vector2([input_array[cleaned_indices[-1]]['X'],input_array[cleaned_indices[-1]]['Z']])-Math.Vector2([input_array[unsorted_indices[0]]['X'],input_array[unsorted_indices[0]]['Z']])
				if(V1.cross(V2)[0]<0):	# so the cavity is on the LEFT
					unsorted_indices.remove(i)
				else:						# so the cavity is on the RIGHT
					for j in range(counter,i):
						counter+=1
						unsorted_indices.remove(j)
					cleaned_indices.append(i)
					unsorted_indices.remove(i)
			counter+=1
		
		# 2: at each discontinuity (one per cavity), add a point that makes the terrain locally vertical (fill the cavity)
		new_cleaned_indices=cleaned_indices[:]
		for cleaned_indice in range(len(cleaned_indices)-1):
			i=cleaned_indices[cleaned_indice]	#the indice of the point before the discontinuity
			j=cleaned_indices[cleaned_indice+1]	#the indice of the point after the discontinuity
			if(j>i+1):	# so there is a discontinuity
				if(input_array[i]['Z']>input_array[j]['Z']): #LEFT side cavity
					line=Geoms.Line(segments_type([input_array[j-1]['X'],input_array[j-1]['Z']],[input_array[j]['X'],input_array[j]['Z']]))
					x=input_array[i]['X']
					z=line.a*x+line.b
					new_cleaned_indices.insert(cleaned_indice+1+(len(new_cleaned_indices)-len(cleaned_indices)),j-1)
					input_array[j-1]["X"]=x
					input_array[j-1]["Z"]=z
				else:	#RIGHT side cavity
					line=Geoms.Line(segments_type([input_array[i]['X'],input_array[i]['Z']],[input_array[i+1]['X'],input_array[i+1]['Z']]))
					x=input_array[j]['X']
					z=line.a*x+line.b
					new_cleaned_indices.insert(cleaned_indice+1+(len(new_cleaned_indices)-len(cleaned_indices)),i+1)
					input_array[i+1]["X"]=x
					input_array[i+1]["Z"]=z
		
		### BEGIN ADD SLOPE TO VERTICAL SEGMENTS
		for indice in range(len(new_cleaned_indices)):
			i=new_cleaned_indices[indice]
			xa=input_array[i]["X"] ; za=input_array[i]["Z"]
			for j in new_cleaned_indices[indice+1:-1]:
				xb=input_array[j]["X"] ; zb=input_array[j]["Z"]
				if(abs(zb-za)<1e-8):	#HORIZONTAL
					break
				neg_slope=(zb-za)<0.
				theta1=np.arctan(abs((xb-xa)/(zb-za)))
				if(neg_slope and theta1<np.radians(1.)):
					input_array[j]["X"]=input_array[i]["X"]+np.tan(np.radians(1.))*(input_array[i]["Z"]-input_array[j]["Z"])
				elif((not neg_slope) and theta1<np.radians(1.) ):
					input_array[j]["X"]=input_array[i]["X"]-np.tan(np.radians(1.))*(input_array[i]["Z"]-input_array[j]["Z"])
				else:
					break
					
		### FINALLY REPLACE THE INPUT_ARRAY
		input_array=input_array[new_cleaned_indices]
		
		### Create segments with X,Z and columns names
		for i in range(0,len(input_array)-1):
			# in all cases, import the terrain coordinates:
			if( abs(input_array['X'][i] - input_array['X'][i+1])<1e-3 and abs(input_array['Z'][i] - input_array['Z'][i+1])<1e-3 ): # consecutive points at the same place
				continue
			s=segments_type([input_array['X'][i],input_array['Z'][i]],[input_array['X'][i+1],input_array['Z'][i+1]])
			for column_name in input_array.dtype.fields:
				name_id = np.where(self.valid_input_attrs[:,0]==column_name)[0]
				if(len(name_id)):
					s.__dict__[column_name]=self.valid_input_attrs[name_id[0],2](input_array[column_name][i])
			self.segments.append(s)
	
	def get_csv_string(self):
		s='X	Z'
		for p in self.valid_input_attrs[:,0]:
			s+="\t"+p
		s=s+"\n"
		for segt in self.segments:
			s+=str(segt.points[0][0])+"\t"+str(segt.points[0][1])+"\t"
			for p in self.valid_input_attrs[:,0]:
				s+=str(segt.__dict__[p])+"\t"
			s=s[:-1]+"\n"
		segt=self.segments[-1]
		s+=str(self.segments[-1].points[1][0])+"\t"+str(self.segments[-1].points[1][1])
		for p in self.valid_input_attrs[:,0]:
			s+=str(segt.__dict__[p])+"\t"
		s=s[:-1]
		return s
	
	def check_segments_continuity(self):
		"""
		Checks the continuity of the terrain. This method checks that each segment starts at the previous segment end. Is the terrain is not valid, the program exits.
		"""
		valid=True
		for i in range(1,len(self.segments)-1):
			if(abs(self.segments[i].points[1,0]-self.segments[i+1].points[0,0])>1e-10):valid=False
			if(abs(self.segments[i].points[1,1]-self.segments[i+1].points[0,1])>1e-10):valid=False
		if(not valid):
			Debug.Print("ERROR, the terrain is not valid")
			sys.exit(1)
	
	def check_segments_parameters_consistency(self):
		"""
			Analyze the segments parameters and checks their consistency/availability. :attr:`forest_available` and :attr:`rebound_models_available` are here.
		"""
		s=self.segments[0] #all the segments has the same data, use the first one to list them below
		HAS_bounce_model_number=s.bounce_model_number is not None
		HAS_roughness=s.roughness is not None
		HAS_R_t=s.R_t is not None
		HAS_trees_density=s.trees_density is not None
		HAS_dhp=(s.dhp_mean is not None) and (s.dhp_std is not None)
		HAS_R_n=s.R_n is not None
		HAS_v_half=s.v_half is not None
		HAS_phi=s.phi is not None
		if(HAS_dhp and HAS_trees_density):
			self.forest_available=True
		if(HAS_bounce_model_number):
			if(HAS_roughness and HAS_R_t and HAS_R_n):
				self.rebound_models_available+=[0,1]
			if(HAS_v_half and HAS_phi):
				self.rebound_models_available+=[2]
				
			bounce_model_numbers=[s.bounce_model_number for s in self.segments]
			USE_classical = 0 in bounce_model_numbers
			USE_pfeiffer = 1 in bounce_model_numbers
			USE_bourrier = 2 in bounce_model_numbers
			if( USE_classical and (0 not in self.rebound_models_available)):
				raise ValueError('At least one segment "rebound_model" parameter has been set to 0(=Classical) but the corresponding parameters were not specified (roughness, R_n, R_t)')
			if( USE_pfeiffer and (1 not in self.rebound_models_available)):
				raise ValueError('At least one segment "rebound_model" parameter has been set to 1(=Pfeiffer) but the corresponding parameters were not specified (roughness, R_n, R_t)')
			if( USE_bourrier and (2 not in self.rebound_models_available)):
				raise ValueError('At least one segment "rebound_model" parameter has been set to 2(=Bourrier) but the corresponding parameters were not specified (roughness, R_t, v_half, phi)')
	
	def compare_params(self,id1,id2):
		s1=self.segments[id1]
		s2=self.segments[id2]
		for param_name in self.valid_input_attrs[:,0]:
			if(s1.__dict__[param_name]!=s2.__dict__[param_name]):
				return False
		return True
		
	def set_params_zones(self):
		if len(self.segments) < 1 :
			self.params_zones=[]
			return
		self.params_zones=[ParamsZone(self,0,0)] #init with first segment = first zone
		for id1 in range(1,len(self.segments)):
			id2 = self.params_zones[-1].end_id
			if(self.compare_params(id1,id2)):	#so we are still on the same param zone, increase its size
				self.params_zones[-1].end_id=id1
			else:	#so we enter a new zone
				self.params_zones.append(ParamsZone(self,id1,id1))
		i=0
		for pz in self.params_zones:
			pz["color"]=cp.qualitative10[i]
			i+=1
	
	def point_is_a_zone_border(self,x):
		id1, id2 = self.get_segments_ids_around_point(x)
		if(id1==-1 or id2==-1): return True
		for pz in self.params_zones:
			if(id1>=pz.start_id and id1<=pz.end_id and id2>=pz.start_id and id2<=pz.end_id):
				return False
		return True

	def get_y(self,x):
		for s in self.segments:
			if(s.points[0,0]<=x and s.points[1][0]>=x):
				return s.get_y(x)
		#x is outside the terrain
		if x>=self.get_points()[-1,0] :
			return self.get_points()[-1,1]
		elif x<=self.get_points()[0,0]:
			return self.get_points()[0,1]
		else:
			Debug.error("Error in terrain.get_y, should never happen.")
	
	def get_segment_zone_ids(self,index):
		for pz in self.params_zones:
			if(index>=pz.start_id and index<=pz.end_id):
				return [pz.start_id,pz.end_id]
				
	def get_points(self):
		"""
		From the list of segments, returns an array of points representing the terrain.
		#FIXME: this method should be used once, but it seems to be used very often during simulation
		
		Returns:
			:class:`numpy.ndarray` [[x0, z0], [x1, z1], ...]
		"""
		output=np.asarray([seg.points[0,:] for seg in self.segments],dtype=float)
		output=np.append(output,[self.segments[-1].points[-1,:]],axis=0)
		return output
	
	def index_segments(self):
		"""
		Index all segments of the terrain, :math:`i=0..N`, applying :class:`Segment` . :attr:`index` :math:`=i`
		As the index are continuously distributed along the terrain, it allows to easily access the next or the previous segment from a given segment.
		"""
		for i in range(0,len(self.segments)):
			self.segments[i].index=i
	
	def get_angle_between(self,id1,id2):
		if(id1>id2):
			id1,id2=id2,id1
		seg1=self.segments[id1]
		seg2=self.segments[id2]
		ret = Math.atan2_signed(seg1.points[0][1]-seg1.points[1][1],seg1.points[0][0]-seg1.points[1][0])- Math.atan2_signed(seg2.points[1][1]-seg2.points[0][1],seg2.points[1][0]-seg2.points[0][0])
		if(ret<0):
			ret+=np.pi*2
		return ret
	
	def get_y_range(self):
		pts=self.get_points()[:,1]
		return [pts.min(), pts.max()]
		
	
	def get_segments_ids_around_point(self,x):
		nearest=np.inf
		for s in self.segments:
			if abs(s.points[0][0]-x)<nearest:
				best_segts = [s.index-1, s.index]
				nearest=abs(s.points[0][0]-x)
		if abs(self.segments[-1].points[1][0]-x)<nearest:
			best_segts = [ len(self.segments)-1, -1]
		return best_segts

	def remove_point(self,x):
		id1, id2 = self.get_segments_ids_around_point(x)
		if(id2==-1): #remove the last point, so remove segment n°id1
			self.remove_segment(id1)
			return
		if id2==len(self.segments)-1 : #remove the penultimate point, so remove the last segment but connect end of id1 to end of id2 before.
			self.segments[id1].points[1][:]=self.segments[id2].points[1][:]
		#general case: remove the segment after the point at x
		self.remove_segment(id2)
	
	def add_point_after(self,x):
		id1, id2 = self.get_segments_ids_around_point(x)
		if(id2!=-1): #if there is a segment after the point
			self.split_segment(id2)
	
	def move_point(self,x,newx,newy):
		id1, id2 = self.get_segments_ids_around_point(x)
		if(id1==-1):
			self.move_segment_point(0,"start",newx,newy)
		elif(id2==-1):
			self.move_segment_point(len(self.segments)-1,"end",newx,newy)
		else:
			self.move_segment_point(id1,"end",newx,newy)
	
	def new_zone_at(self,x):
		id1, id2 = self.get_segments_ids_around_point(x)
		if not self.point_is_a_zone_border(x) :
			zone_start, zone_end = self.get_segment_zone_ids(id2)
			for index in range(id2,zone_end+1):
				for p in self.valid_input_attrs:
					self.segments[index].__dict__[p[0]]=p[3]
		self.set_params_zones()

	def remove_segment(self,id):
		if(id!=0 and id!=len(self.segments)-1): #if the segment is not at an extremity, update the end point of the previous segment.
			self.segments[id-1].points[1][:]=self.segments[id].points[1][:]
			self.segments[id-1].set_geometrical_quantities()
		self.segments.pop(id)
		self.index_segments()
		self.set_params_zones()
		self.check_segments_continuity()
		self.check_segments_parameters_consistency()
	
	def split_segment(self,id):
		middle_point = (self.segments[id].points[0] + self.segments[id].points[1])/2
		new_segt=copy.deepcopy(self.segments[id])
		self.segments[id].points[1]=middle_point[:]
		self.segments[id].set_geometrical_quantities()
		new_segt.points[0]=middle_point[:]
		new_segt.set_geometrical_quantities()
		self.segments.insert(id+1,new_segt)
		self.index_segments()
		self.set_params_zones()
		self.check_segments_continuity()
		self.check_segments_parameters_consistency()
	
	def move_segment_point(self,index,which_end,x,y):
		eps=1e-4
		if(which_end=="start"):
			x=min(x,self.segments[index].points[1][0]-eps)
			if(index!=0):
				x=max(x,self.segments[index-1].points[0][0]+eps)
			self.segments[index].points[0][0]=x
			self.segments[index].points[0][1]=y
			self.segments[index].set_geometrical_quantities()
			if(index!=0):
				self.segments[index-1].points[1][0]=x
				self.segments[index-1].points[1][1]=y
				self.segments[index-1].set_geometrical_quantities()
		elif(which_end=="end"):
			x=max(x,self.segments[index].points[0][0]+eps)
			if(index!=len(self.segments)-1):
				x=min(x,self.segments[index+1].points[1][0]-eps)
			self.segments[index].points[1][0]=x
			self.segments[index].points[1][1]=y
			self.segments[index].set_geometrical_quantities()
			if(index!=len(self.segments)-1):
				self.segments[index+1].points[0][0]=x
				self.segments[index+1].points[0][1]=y
				self.segments[index+1].set_geometrical_quantities()
		self.index_segments()
		self.set_params_zones()
		self.check_segments_continuity()
		self.check_segments_parameters_consistency()

class ParamsDict(dict): #subclass dict to be able to i)put attributes (with ".") on a dict ; and ii) propagate params to segments on key assignation (with "[]").
	def __init__(self,zone,terrain):
		super(ParamsDict,self).__init__()
		self.zone=zone
		for attr in terrain.valid_input_attrs[:,0]: #import parameters from start_id
			super(ParamsDict,self).__setitem__(attr, self.zone.segments[self.zone.start_id].__dict__[attr]) #use "super" to avoid the usage of __setitem__ below.

	def __setitem__(self,item,value):
		super(ParamsDict,self).__setitem__(item,value)
		for i in range(self.zone.start_id, self.zone.end_id+1):
			self.zone.segments[i].__dict__[item]=value

class ParamsZone(dict):
	def __init__(self,terrain,start,end):
		self.start_id=start
		self.end_id=end
		self.segments=terrain.segments
		self["color"]=[0,0,0]
		self["params"]=ParamsDict(self,terrain)
		
class Checkpoint(object):
	"""
	A checkpoint along the terrain. Checkpoints data are post-computed from all contacts of all rocks.
	
	Args:
		x (float): the x-position of the checkpoint
		base_height (float): the height of the terrain at the checkpoint x
		rocks (list [:class:`Rock`, ...]): all rocks that passed through the checkpoint
		heights (list [float, ...]): all heights at which the rocks passed through the checkpoint
		vels (list [[vx1,vz1], [vx2,vz2], ...]): all velocities at which the rocks passed through the checkpoint
		angVels (list [float, ...]): all angVels at which the rocks passed through the checkpoint
	"""
	def __init__(self,x):
		self.x=x
		#self.rocks=[]	#this will avoid error when looping on empty checkpoints
	def init_data(self,simulation):
		"""
		Initialize data lists: :attr:`rocks`, :attr:`heights`, :attr:`vels`, :attr:`angVels`
		"""
		self.base_height=simulation.terrain.get_y(self.x)
		#self.rocks=[]
		self.heights=[]
		self.vels=[]
		self.angVels=[]
		self.rocks_ids=[]













"""
Copyright (c) 2017, 2018, 2019 Irstea
Copyright (c) 2017, 2018, 2019 François Kneib
Copyright (c) 2017, 2018, 2019 Franck Bourrier
Copyright (c) 2017, 2018, 2019 David Toe
Copyright (c) 2017, 2018, 2019 Frédéric Berger
Copyright (c) 2017, 2018, 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""

# TODO: forest integration 
import numpy as np
#from . import Objects
import platrock.Common.Math as Math
from . import RasterTools
import platrock.Common.Outputs as Out


class Postprocessing(object):
	def __init__(self,simulation,raster_cell_length=None):
		self.simulation=simulation
		if(raster_cell_length is None):
			self.raster=self.simulation.terrain.Z_raster
		else:
			self.raster=RasterTools.from_raster(simulation.terrain.Z_raster,cell_length=float(raster_cell_length))
	
	def get_indices_cleaned(self,arr):
		arr_shift1=np.roll(arr,1,axis=0) ; arr_shift1[0]=np.NaN
		arr_shift2=np.roll(arr,-1,axis=0); arr_shift2[-1]=np.NaN
		XCompare=np.where(arr_shift1[:,0]!=arr_shift2[:,0])[0]
		YCompare=np.where(arr_shift1[:,1]!=arr_shift2[:,1])[0]
		return np.unique( np.concatenate((XCompare,YCompare)) )
	
	def get_raster_indices_from_contacts(self,c_pos,c_types):
		Len=np.sum(c_types!=Out.OUT) # exclude out of bounds contacts as it will likely give cells indices outside raster.
		raster_indices=np.empty([Len,2],dtype=int)
		for i in range(Len):
			raster_indices[i]=self.raster.get_indices_from_coords(c_pos[i,0:2])
		return raster_indices
	
	def insert_fly(self,arrays_dict,where):
		for k in arrays_dict.keys():
			arrays_dict[k]=np.insert(arrays_dict[k],where,np.zeros(arrays_dict[k][0].shape),axis=0)
		arrays_dict["pos"][where]=arrays_dict["pos"][where-1]	#in the case of a flight, pos and vel are copied from the last known contact
		arrays_dict["vels"][where]=arrays_dict["vels"][where-1]
		arrays_dict["types"][where]=Out.FLY
		arrays_dict["angVels"][where]=arrays_dict["angVels"][where-1]
			
	def add_flights_to_raster_cells_and_contacts(self, raster_indices, all_c_fields):
		# A new contact type is now added : flights_over_cells
		#DETECT FLYING (contactless) CELLS ON THE RASTER :
		i=0
		while i < len(raster_indices)-1:
			ix, iy = raster_indices[i,0], raster_indices[i,1]
			ix_next, iy_next=raster_indices[i+1,0], raster_indices[i+1,1]
			cells_indices_distance=abs(ix_next-ix) + abs(iy_next-iy) # -> how far are the cells if I only jump through the cells edges (no diagonals) ?
			if(cells_indices_distance<=1): # consecutive contacts on the same raster cell (=0), or consecutive contacts on neighbours raster cell (=1)
				i+=1
				continue
			else:	#a fly occured : complete the rock path on the raster with flying cells
				current_pos=all_c_fields["pos"][i][0:2]
				branch=all_c_fields["pos"][i+1][0:2]-current_pos
				branch_angle=Math.atan2_unsigned(branch[1],branch[0])	#NOTE : x and y are inverted, this is a convention (see https://en.wikipedia.org/wiki/Atan2#Illustrations)
				ix2=ix ; iy2=iy
				while(abs(ix_next-ix2) + abs(iy_next-iy2)>1): #while we didn't reached a neighbour of the next contact cell ...
					SE_branch=np.array([ self.raster.X[ix2+1] , self.raster.Y[iy2] ])-current_pos#South-East
					SE_angle=Math.atan2_unsigned(SE_branch[1],SE_branch[0])
					
					SW_branch=np.array([ self.raster.X[ix2] , self.raster.Y[iy2] ])-current_pos#South-West
					SW_angle=Math.atan2_unsigned(SW_branch[1],SW_branch[0])
					
					if(branch_angle<=SE_angle and branch_angle>=SW_angle): #exiting the current cell by the BOTTOM edge
						current_pos=current_pos+branch*((self.raster.Y[iy]-current_pos[1])/branch[1])
						iy2-=1
						raster_indices=np.insert(raster_indices,i+1,[ix2,iy2],axis=0)
						self.insert_fly(all_c_fields,i+1)
						i+=1
					else:
						NW_branch=np.array([ self.raster.X[ix2] , self.raster.Y[iy2+1] ])-current_pos#North-West
						NW_angle=Math.atan2_unsigned(NW_branch[1],NW_branch[0])
						
						if(branch_angle<SW_angle and branch_angle>=NW_angle):#exiting the current cell by the LEFT edge
							current_pos=current_pos+branch*((self.raster.X[ix]-current_pos[0])/branch[0])
							ix2-=1
							raster_indices=np.insert(raster_indices,i+1,[ix2,iy2],axis=0)
							self.insert_fly(all_c_fields,i+1)
							i+=1
						else:
							NE_branch=np.array([ self.raster.X[ix2+1] , self.raster.Y[iy2+1] ])-current_pos#North-West
							NE_angle=Math.atan2_unsigned(NE_branch[1],NE_branch[0])
							if(branch_angle<NW_angle and branch_angle>=NE_angle):#exiting the current cell by the TOP edge
								current_pos=current_pos+branch*((self.raster.Y[iy+1]-current_pos[1])/branch[1])
								iy2+=1
								raster_indices=np.insert(raster_indices,i+1,[ix2,iy2],axis=0)
								self.insert_fly(all_c_fields,i+1)
								i+=1
							else:#exiting the current cell by the RIGHT edge
								current_pos=current_pos+branch*((self.raster.X[ix+1]-current_pos[0])/branch[0])
								ix2+=1
								raster_indices=np.insert(raster_indices,i+1,[ix2,iy2],axis=0)
								self.insert_fly(all_c_fields,i+1)
								i+=1
				i+=1
		return (raster_indices,all_c_fields)
	
	def run(self):
		self.raster.add_data_grid("crossings",int,0)
		self.raster.add_data_grid("heights",list)
		self.raster.add_data_grid("vels",list)
		self.raster.add_data_grid("stops",list) #at each stop, store the START cell coordinates
		self.raster.add_data_grid("Ec",list) #Kinetic energy
		for r_nb in range(self.simulation.nb_rocks):
			
			c_pos   = self.simulation.output.get_contacts_pos(r_nb)
			c_types = self.simulation.output.get_contacts_types(r_nb)
			
			#find cells from contacts pos:
			raster_indices = self.get_raster_indices_from_contacts(c_pos,c_types)
			
			#CLEANING PASS : when consecutive contacts on the same raster cell occurs, only keep the first one and the last one
			indices_duplicates_removed=self.get_indices_cleaned(raster_indices)
			raster_indices=raster_indices[indices_duplicates_removed]	# LIST OF RASTERS ON WHICH CONTACTS OCCURED
			c_pos=c_pos[indices_duplicates_removed]
			c_vels=self.simulation.output.get_contacts_vels(r_nb)[indices_duplicates_removed]
			c_types=c_types[indices_duplicates_removed]
			c_angVels=self.simulation.output.get_contacts_angVels(r_nb)[indices_duplicates_removed]
			all_c_fields={"pos":c_pos,"vels":c_vels,"types":c_types,"angVels":c_angVels}
			
			# add flights (cells with no contacts). Finally, len(raster_indices) == len(contacts), allowing to loop over both at the same time.
			raster_indices, all_c_fields = self.add_flights_to_raster_cells_and_contacts(raster_indices, all_c_fields)
			
			# Count the number of rocks that passed over each cell :
			rock_count=np.zeros(np.shape(self.raster.data["crossings"]),dtype=int) #init to 0
			for index in raster_indices:	#loop on cells the rocks passed over
				rock_count[index[0],index[1]]+=1	#increment
			rock_count[rock_count>0]=1			#limit the value to 0 to avoid multiple rebounds on a single cell
			self.raster.data["crossings"]+=rock_count	#add this rock_count to the global data
			
			self.raster.data["heights"][raster_indices[0,0],raster_indices[0,1]].append(all_c_fields["pos"][0][2])
			self.raster.data["vels"][raster_indices[0,0],raster_indices[0,1]].append(np.linalg.norm(all_c_fields["vels"][0]))
			self.raster.data["Ec"][raster_indices[0,0],raster_indices[0,1]].append(
				0.5*(
					self.simulation.output.densities[r_nb]*self.simulation.output.volumes[r_nb]*np.linalg.norm(all_c_fields["vels"][0])**2 + #translation
					np.dot(all_c_fields["angVels"][0],np.dot(self.simulation.output.inertias[r_nb],all_c_fields["angVels"][0])) #rotation
				)
			) # See below to have a more explicit computation of Ec
			for i in range(1,len(raster_indices)): #FIXME : if a rock come back into a cell, avoid it
				index=raster_indices[i]
				prev_index=raster_indices[i-1]
				if((index==prev_index).all()): # if the previous contact was in the same cell, do nothing
					continue
				previous_c_pos=all_c_fields["pos"][i-1]
				previous_c_vel=all_c_fields["vels"][i-1]
				if(prev_index[1]-index[1] == 1):# the rock came to the current cell from the north face
					yM=self.raster.Y[prev_index[1]]	#the y coordinate of north face of the cell
					time=(yM-previous_c_pos[1])/previous_c_vel[1] # time = flight time from last effective contact to the entering of the rock in the cell
					xM=previous_c_vel[0]*time+previous_c_pos[0]
				elif(prev_index[1]-index[1] == -1): # the rock came to the current cell from the south face
					yM=self.raster.Y[index[1]]	#the y coordinate of south face of the cell
					time=(yM-previous_c_pos[1])/previous_c_vel[1]
					xM=previous_c_vel[0]*time+previous_c_pos[0]
				elif(prev_index[0]-index[0] == 1):	# the rock came to the current cell from the east face
					xM=self.raster.X[prev_index[0]] #the x coordinate of east face of the cell
					time=(xM-previous_c_pos[0])/previous_c_vel[0]
					yM=previous_c_vel[1]*time+previous_c_pos[1]
				elif(prev_index[0]-index[0] == -1): # the rock came to the current cell from the west face
					xM=self.raster.X[index[0]] #the x coordinate of the west face of the cell
					time=(xM-previous_c_pos[0])/previous_c_vel[0]
					yM=previous_c_vel[1]*time+previous_c_pos[1]
				if(np.isnan(time)):	# NOTE: this is for a very specific case, if the rebound occurs exactly on a cell edge and with vx=0 and/or vy=0.
					vel=Math.Vector3(all_c_fields["vels"][i])
					absolute_height=all_c_fields["pos"][i][2]
				else: # this is the general case
					absolute_height=-0.5*self.simulation.gravity*time**2 + previous_c_vel[2]*time + previous_c_pos[2] #FIXME : compute relative height
					vel=Math.Vector3([previous_c_vel[0], previous_c_vel[1], previous_c_vel[2] - self.simulation.gravity*time])
				
				self.raster.data["heights"][index[0],index[1]].append(absolute_height)
				self.raster.data["vels"][index[0],index[1]].append(vel.norm())
				#Compute Ec and add it the the raster cell:
				I=self.simulation.output.inertias[r_nb]
				angVel=all_c_fields["angVels"][i]
				density=self.simulation.output.densities[r_nb]
				volume=self.simulation.output.volumes[r_nb]
				Ec=0.5*( density*volume*vel.norm()**2 + np.dot(angVel,np.dot(I,angVel)) ) # mv² + ω_t X I X ω
				self.raster.data["Ec"][index[0],index[1]].append(Ec)
			
			#Handle stops:
			if(c_types[-1]==Out.STOP):
				self.raster.data["stops"][index[0],index[1]].append(raster_indices[0])
		
		#Compute stats rasters based on other rasters:
		ids=np.where(self.raster.data["crossings"]!=0)
		for field in ["heights","vels", "Ec"]:
			for quantile in [90,95,99]:
				raster_data=self.raster.add_data_grid(field+"_quantile-"+str(quantile)+"%",float)
				for i,j in zip(*ids):
					raster_data[i,j]=np.quantile(self.raster.data[field][i,j],quantile/100,interpolation="linear")
		
		raster_data=self.raster.add_data_grid("number_of_source-cells",int)
		for i in range(self.raster.nx):
			for j in range(self.raster.ny):
				Len=len(self.raster.data["stops"][i,j])
				if Len>0:
					raster_data[i,j]=len(np.unique(self.raster.data["stops"][i,j],axis=0))
	
	def plot(self):
		import platrock.GUI.Plot3D
		self.plot=platrock.GUI.Plot3D



"""
Copyright (c) 2019 François Kneib
Copyright (c) 2019 Franck Bourrier
Copyright (c) 2019 David Toe
Copyright (c) 2019 Frédéric Berger
Copyright (c) 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""

class FriendImporter(object):
	friendClass=None
	friendImportNames=[]
	
	#Import normal ATTRIBUTES, i.e. the ones handeled by the objects.
	def __init__(self,*args,**kwargs):
		friendObject=self.friendClass(*args,**kwargs)
		for name in self.friendImportNames:
			if name in friendObject.__dict__.keys(): # so "name" is handeled by the object (like attributes)
				self.__dict__[name]=friendObject.__dict__[name]
	
	#Import METHOD and CLASS ATTRIBUTES that are handeled by the class, not the objects.
	#This is called at PlatRock launch early state, when FriendImporter class is subclassed (not sub-instanciated).
	def __init_subclass__(child_class):
		for name in child_class.friendImportNames:
			if name in child_class.friendClass.__dict__.keys(): #so "name" is handeled by the class (like methods and class attributes)
				exec('child_class.'+name+'=child_class.friendClass.'+name)



"""
Copyright (c) 2017, 2018, 2019 Irstea
Copyright (c) 2017, 2018, 2019 François Kneib
Copyright (c) 2017, 2018, 2019 Franck Bourrier
Copyright (c) 2017, 2018, 2019 David Toe
Copyright (c) 2017, 2018, 2019 Frédéric Berger
Copyright (c) 2017, 2018, 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import matplotlib as mpl
import platrock
from descartes import PolygonPatch
import plotly.offline as plyo
import plotly.graph_objs as plygo
import platrock.Common.ColorPalettes as cp

import matplotlib as mpl
if platrock.web_ui:
	mpl.use('Agg')
	#import plotly as ply
import matplotlib.pyplot as plt
if not platrock.web_ui: plt.ion()

mpl.rcParams["figure.subplot.bottom"]=0.05
mpl.rcParams["figure.subplot.top"]=0.95
mpl.rcParams["figure.subplot.left"]=0.0
mpl.rcParams["figure.subplot.right"]=0.92

fig=plt.figure("PlatRock Raster View",facecolor="w",edgecolor="w")
ax=plt.axes()

def clear():
	global fig,ax
	plt.close("all")
	fig=plt.figure("PlatRock Raster View",facecolor="w",edgecolor="w")
	ax=plt.axes()


def draw_terrain(s, with_polygons=False):
	terrain=s.terrain
	
	#1- background raster:
	azimuth=315
	angle_altitude=45
	x, y = np.gradient(terrain.Z_raster.data["Z"].transpose())
	X,Y=np.meshgrid(terrain.Z_raster.X, terrain.Z_raster.Y)
	slope = np.pi/2. - np.arctan(np.sqrt(x*x + y*y))
	aspect = np.arctan2(-x, y)
	azimuthrad = azimuth*np.pi / 180.
	altituderad = angle_altitude*np.pi / 180.
	shaded = np.sin(altituderad) * np.sin(slope)\
	+ np.cos(altituderad) * np.cos(slope)\
	* np.cos(azimuthrad - aspect)
	plt.pcolormesh(X,Y,255*(shaded + 1)/2,cmap='Greys',rasterized=True)
	
	if(with_polygons):
		#4- terrain polygons
		for soil_param_set in terrain.soil_params:
			ax.add_patch(PolygonPatch(soil_param_set["shapely_polygon"],facecolor=np.append(soil_param_set["color"]/255,0.65),edgecolor=soil_param_set["color"][0:3]/255,linewidth=1.5))
		
		#3- rocks start positions:
		for start_param_set in s.rocks_start_params:
			ax.add_patch(PolygonPatch(start_param_set["shapely_polygon"],facecolor=np.append(start_param_set["color"]/255,0.65),edgecolor="black",linewidth=1))
	
	fig.set_size_inches(3+1,3*terrain.Z_raster.data["Z"].shape[1]/terrain.Z_raster.data["Z"].shape[0])
	ax.set_aspect("equal")
	fig.canvas.draw()
	

def draw_forest(terrain):
	ax.plot(terrain.trees_as_array[:,0],terrain.trees_as_array[:,1],"o",color="green")
	fig.canvas.draw()
	
def draw_rocks_count(raster):
	data=raster.data['crossings']
	#use masked array to set a "bad" value to the cmap. It allows to set the cmap to fully transparent when crossings==0
	masked_array = np.ma.array (data, mask=(data==0))
	cmap = mpl.cm.jet
	cmap.set_bad('gray',0.)	#gray color but opacity=0
	x,y=np.meshgrid(raster.X, raster.Y,indexing="ij")
	cm=ax.pcolormesh(x,y,masked_array,cmap=cmap,edgecolor=[0.6,0.6,0.8],linewidth=0.01,zorder=1)
	ax.set_title("Nombre de passages")
	cbar=plt.colorbar(cm,aspect=30)
	cbar.set_label("# passages")
	ax.set_aspect("equal")
	fig.canvas.draw()

def draw_mean_vel(raster):
	data=raster.data_to_mean('vels')
	#use masked array to set a "bad" value to the cmap. It allows to set the cmap to fully transparent when vel==0
	masked_array = np.ma.array (data, mask=(data==0))
	cmap = mpl.cm.jet
	cmap.set_bad('gray',0.)	#gray color but opacity=0
	x,y=np.meshgrid(raster.X, raster.Y,indexing="ij")
	cm=ax.pcolormesh(x,y,masked_array,cmap=cmap,edgecolor=[0.6,0.6,0.8],linewidth=0.01,zorder=1)
	ax.set_title("Vitesses moyennes")
	cbar=plt.colorbar(cm,aspect=30)
	cbar.set_label("$[m.s^{-1}]$")
	ax.set_aspect("equal")
	fig.canvas.draw()

def draw_mean_Ec(raster):
	data=raster.data_to_mean('Ec')/1000.
	#use masked array to set a "bad" value to the cmap. It allows to set the cmap to fully transparent when vel==0
	masked_array = np.ma.array (data, mask=(data==0))
	cmap = mpl.cm.jet
	cmap.set_bad('gray',0.)	#gray color but opacity=0
	x,y=np.meshgrid(raster.X, raster.Y,indexing="ij")
	cm=ax.pcolormesh(x,y,masked_array,cmap=cmap,edgecolor=[0.6,0.6,0.8],linewidth=0.01,zorder=1)
	ax.set_title("Énergie cinétique moyenne")
	cbar=plt.colorbar(cm,aspect=30)
	cbar.set_label("$[kJ]$")
	ax.set_aspect("equal")
	fig.canvas.draw()

def draw_number_of_source_cells(raster):
	data=raster.data['number_of_source-cells']
	#use masked array to set a "bad" value to the cmap. It allows to set the cmap to fully transparent when vel==0
	masked_array = np.ma.array (data, mask=(data==0))
	cmap = mpl.cm.jet
	cmap.set_bad('gray',0.)	#gray color but opacity=0
	x,y=np.meshgrid(raster.X, raster.Y,indexing="ij")
	cm=ax.pcolormesh(x,y,masked_array,cmap=cmap,edgecolor=[0.6,0.6,0.8],linewidth=0.01,zorder=1)
	ax.set_title("Nombre de cellules source")
	cbar=plt.colorbar(cm,aspect=30)
	cbar.set_label("$Number$")
	ax.set_aspect("equal")
	fig.canvas.draw()

def draw_rocks_trajectories(s,rocks_ids):
	for rock_nb in rocks_ids:
		draw_rock_trajectory(s,rock_nb,draw=False)
	fig.canvas.draw()
			
def draw_rock_trajectory(s,nb,draw=True):
	pos=s.output.get_contacts_pos(nb)
	ax.plot(pos[:,0],pos[:,1],'-o',ms=0.5,lw=1,zorder=99)
	if(draw):
		fig.canvas.draw()

def draw_sample_trajectories(s,nb=50):
	rocks_to_plot=[]
	if(nb>=s.nb_rocks):
		rocks_to_plot=range(s.nb_rocks)
	else:
		rocks_to_plot=np.linspace(0,s.nb_rocks-1,nb,dtype=int)
	draw_rocks_trajectories(s,rocks_to_plot)

def get_plotly_raw_html(s, sample_trajectories=False):
	#Initialize the lists
	data = []
	annotations = []
	
	x=[]
	y=[]
	z=[]
	I=[]
	J=[]
	K=[]
	s.before_run_tasks()
	for p in s.terrain.points:
		x.append(p.pos[0])
		y.append(p.pos[1])
		z.append(p.pos[2])
	for f in s.terrain.faces:
		I.append(f.points[0].id-1)
		J.append(f.points[1].id-1)
		K.append(f.points[1].id-1)
		
	triangles=plygo.Mesh3d(
		x=x,
		y=y,
		z=z,
		hoverinfo='skip'
		#facecolor="rgb(255, 0,0)",
		#i=I,
		#j=J,
		#k=K,
		#name=''
	)
	data.append(triangles)
	axis = dict(
		showbackground=True,
		backgroundcolor="rgb(230, 230,230)",
		gridcolor="rgb(255, 255, 255)",
		zerolinecolor="rgb(255, 255, 255)",
    )
	#4- Configure layout (axes, etc...):
	layout = plygo.Layout(
		autosize=False,
		width=1000,
		height=800,
		margin=plygo.layout.Margin(
			l=40,
			r=10,
			b=20,
			t=30,
			pad=4
		),
		scene=dict(
			xaxis=dict(axis),
			yaxis=dict(axis),
			zaxis=dict(axis),
			#aspectratio=dict(
				#x=1,
				#y=1,
				#z=0.5
			#)
		),
		showlegend= False,
		#hovermode = 'closest',
		#annotations=annotations,
	)
	
	#5- configure plotly menubar:
	config = dict(
		modeBarButtonsToRemove = ["autoScale2d","toggleSpikelines","hoverCompareCartesian","hoverClosestCartesian","select2d","lasso2d"],
		displaylogo = False,
		displayModeBar = True,
	)
	
	#6- Generate plotly html/js:
	ply_fig = plygo.Figure(data=data, layout=layout)
	return plyo.plot(ply_fig, config=config, show_link=False, output_type="div", include_plotlyjs=False)







"""
Copyright (c) 2017, 2018, 2019 Irstea
Copyright (c) 2017, 2018, 2019 François Kneib
Copyright (c) 2017, 2018, 2019 Franck Bourrier
Copyright (c) 2017, 2018, 2019 David Toe
Copyright (c) 2017, 2018, 2019 Frédéric Berger
Copyright (c) 2017, 2018, 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""

#cython: language_level=3, boundscheck=False, wraparound=False
#add ", profile=True" to the above line to activate cython profiling.

from numpy cimport ndarray
cimport numpy as np
cimport cython

from libc.math cimport sqrt
import platrock.ThreeD.Objects as Objects

cdef double squared_norm3(double* A):
	return A[0]**2+A[1]**2+A[2]**2

cdef double norm3(double* A):
	return sqrt(squared_norm3(A))

cdef double dot3(double* A, double* B):
	cdef double result
	result = A[0]*B[0]+A[1]*B[1]+A[2]*B[2]
	return result

#NOTE: commented as unused
#cdef double dot2(double* A, double* B):
	#cdef double result
	#result = A[0]*B[0]+A[1]*B[1]
	#return result

cdef double* cross3(double* A, double* B, double* out):
	out[0]=A[1]*B[2]-A[2]*B[1]
	out[1]=A[2]*B[0]-A[0]*B[2]
	out[2]=A[0]*B[1]-A[1]*B[0]
	return out

cdef double cross2(double* A, double *B):
	return A[0]*B[1]-A[1]*B[0]

cdef double* sub2(double* A, double* B, double* out):
	for i in range(2):
		out[i]=A[i]-B[i]
	return out

cdef double* sub3(double* A, double* B, double* out):
	for i in range(3):
		out[i]=A[i]-B[i]
	return out

cdef double* add3(double* A, double* B, double* out):
	for i in range(3):
		out[i]=A[i]+B[i]
	return out

cdef double* mult_s3(double A, double* B, double* out):
	for i in range(3):
		out[i]=A*B[i]
	return out

cdef bint is_point_inside_3D(double* p0, double* p1, double* p2, double* n ,double* p):
	cdef double[3] point_on_face
	sub3(p, mult_s3(dot3(sub3(p,p0,point_on_face),n),n,point_on_face),point_on_face)	#compute point_on_face
	cdef double[3] buff_1, buff_2, buff_3
	cdef double cp1 = dot3( cross3(sub3(p1,p0,buff_1), sub3(point_on_face,p0,buff_2),buff_3), n)
	cdef double cp2 = dot3( cross3(sub3(p2,p1,buff_1), sub3(point_on_face,p1,buff_2),buff_3), n)
	cdef double cp3 = dot3( cross3(sub3(p0,p2,buff_1), sub3(point_on_face,p2,buff_2),buff_3), n)
	cdef bint result = ((cp1>0 and cp2>0 and cp3>0) or (cp1<0 and cp2<0 and cp3<0))
	return result

cdef bint is_point_inside_2D(double* p0, double* p1, double* p2, double* n ,double* p):
	cdef double[2] point_xy = [p[0],p[1]]
	cdef double[2] buff_1, buff_2
	cdef double cp1 = cross2(sub2(p1,p0,buff_1), sub2(point_xy,p0,buff_2))
	cdef double cp2 = cross2(sub2(p2,p1,buff_1), sub2(point_xy,p1,buff_2))
	cdef double cp3 = cross2(sub2(p0,p2,buff_1), sub2(point_xy,p2,buff_2))
	cdef bint result = ((cp1>0 and cp2>0 and cp3>0) or (cp1<0 and cp2<0 and cp3<0))
	return result

def verlet_run(Verlet_update_instance,s):
	r=s.current_rock
	
	cdef double r_radius = r.radius
	cdef double bb_half_length = r.bounding_box.half_length
	cdef double[3] r_pos = r.pos
	cdef double[3] bb_pos = r.bounding_box.pos
	cdef int i #counters
	
	#1-DO WE HAVE TO UPDATE THE VERLET LIST ?
	if( not(
			(r_pos[0]-r_radius*1.1<bb_pos[0]-bb_half_length) or
			(r_pos[1]-r_radius*1.1<bb_pos[1]-bb_half_length) or
			(r_pos[2]-r_radius*1.1<bb_pos[2]-bb_half_length) or
			(r_pos[0]+r_radius*1.1>bb_pos[0]+bb_half_length) or
			(r_pos[1]+r_radius*1.1>bb_pos[1]+bb_half_length) or
			(r_pos[2]+r_radius*1.1>bb_pos[2]+bb_half_length)
		)
	):return

	
	#2-UPDATE THE BOUNDING BOX PROPERTIES
	r.bounding_box.half_length=Verlet_update_instance.dist_factor*r.radius
	bb_half_length = r.bounding_box.half_length
	r.bounding_box.pos[:]=r.pos[:]
	bb_pos=r.bounding_box.pos
	
	#3-GET THE FACES OF TERRAIN THAT HAVE AT LEAST ONE POINT IN THE BOUNDING SPHERE
	cdef int size = s.terrain.faces_xyz_bb.shape[0]
	cdef double[:,:] faces_bb_min_max_mv = s.terrain.faces_xyz_bb
	r.verlet_faces_list=[]
	for i in range(size):
		outside=(	(bb_pos[0]+bb_half_length<faces_bb_min_max_mv[i,0]) or
					(bb_pos[1]+bb_half_length<faces_bb_min_max_mv[i,2]) or
					(bb_pos[2]+bb_half_length<faces_bb_min_max_mv[i,4]) or
					(bb_pos[0]-bb_half_length>faces_bb_min_max_mv[i,1]) or
					(bb_pos[1]-bb_half_length>faces_bb_min_max_mv[i,3]) or
					(bb_pos[2]-bb_half_length>faces_bb_min_max_mv[i,5])
				)
		if(not outside):r.verlet_faces_list.append(s.terrain.faces[i])
	
	#4-GET THE TREES OF TERRAIN THAT ARE IN THE BOUNDING SPHERE
	cdef double[:,:] trees_mv
	if(s.enable_forest):
		trees_mv = s.terrain.trees_as_array
		size = s.terrain.trees_as_array.shape[0]
		r.verlet_trees_list=[]
		for i in range(size):
			if( (trees_mv[i,0]-r_pos[0])**2 + (trees_mv[i,1]-r_pos[1])**2 <  (bb_half_length - trees_mv[i,2]/2)**2 ):
				r.verlet_trees_list.append(s.terrain.trees[i])


def contacts_detector_run(self,s):	#for sphere only !
		
		# ROCK-TERRAIN :
		r=s.current_rock
		r.terrain_active_contact = None
		cdef double r_radius = r.radius
		cdef double face_candidate_dist = 1e10
		cdef double[3] r_pos = r.pos
		cdef double[3] r_vel = r.vel
		cdef double[3] face_normal, branch, edge, proj
		cdef double[3][3] face_pts
		cdef double[3] buff
		cdef double proj_coef
		cdef int nb_faces = len(r.verlet_faces_list)
		cdef int pi,i,fi #counters
		cdef int face_contact_indice = -1
		cdef double dist
		cdef bint found_edge_contact = False
		
		for fi in range(nb_faces):
			for i in range(3):
				face_pts[i] = r.verlet_faces_list[fi].points[i].pos
			face_normal = r.verlet_faces_list[fi].normal
			dist=dot3( face_normal,sub3( r_pos,face_pts[0],buff ) )-r_radius
			if(dist>0): # no penetration with the plan : no contact
				continue
			#if(dot3(r_vel,face_normal)>0):
				#continue
			if(is_point_inside_3D(face_pts[0], face_pts[1], face_pts[2], face_normal,r_pos)):
				if(dist<face_candidate_dist):
					face_contact_indice=fi
					face_candidate_dist=dist
				continue
			found_edge_contact=False
			for pi in range(-1,2):
				branch=sub3(r_pos,face_pts[pi],buff)
				edge=sub3(face_pts[pi+1],face_pts[pi],buff)
				proj_coef=dot3(branch,edge)/squared_norm3(branch)
				if(proj_coef>1 or proj_coef<0):continue
				proj=mult_s3(proj_coef,edge,buff)
				dist=norm3(sub3(branch,proj,buff))-r_radius
				if(dist<face_candidate_dist):
					face_contact_indice=fi
					face_candidate_dist=dist
					found_edge_contact=1
			if(found_edge_contact):continue
			for pi in range(3):
				dist=norm3(sub3(face_pts[pi],r_pos,buff))-r_radius
				if(dist<face_candidate_dist):
					#this is currently the best candidate
					face_contact_indice=fi
					face_candidate_dist=dist
		if(face_candidate_dist<0):
			r.terrain_active_contact = Objects.Rock_terrain_contact(r.points[0],r.verlet_faces_list[face_contact_indice])
			r.terrain_active_contact.dist=face_candidate_dist
		
		# ROCK - TREES :
		cdef double[2] tree_pos
		cdef double tree_dhp
		cdef bint tree_active
		if(s.enable_forest):
			r.tree_active_contact=None
			size = len(r.verlet_trees_list)
			for fi in range(size):
				tree_active = r.verlet_trees_list[fi].active
				if(tree_active):
					tree_pos = r.verlet_trees_list[fi].pos
					tree_dhp = r.verlet_trees_list[fi].dhp
					dist=(tree_pos[0]-r_pos[0])**2 + (tree_pos[1]-r_pos[1])**2	#NOTE: this is not the real dist
					if( dist < (r_radius+tree_dhp/2/100)**2 ):
						r.tree_active_contact=Objects.Rock_tree_contact(r.verlet_trees_list[fi])
						r.tree_active_contact.dist=sqrt(dist)-r_radius+tree_dhp/2

def nscd_integrator_run(self,s):
	r=s.current_rock
	cdef double[:] vel = s.current_rock.vel
	cdef double dt = s.dt
	cdef double gravity = s.gravity
	vel[2]-=gravity*dt
	cdef double[:] pos = s.current_rock.pos
	for i in range(3):
		pos[i]+=vel[i]*dt
	if(r.pos[2]<s.terrain.min_height):
		r.is_stopped=True
		r.out_of_bounds=True
	
from platrock.Common import  Math
def set_faces_normals_and_cogs(faces):
	cdef int nb_faces = len(faces)
	cdef int i,j,k
	cdef double[3] normal, cog, buff1, buff2, buff3
	cdef double[3][3] points
	for i in range(nb_faces):
		for j in range(3):
			for k in range(3):
				points[j][k]=faces[i].points[j].pos[k]
		normal=cross3(sub3(points[0],points[1],buff1),sub3(points[2],points[1],buff2),buff3)
		normal=mult_s3(1./norm3(normal),normal,buff1)
		if(normal[2]<0.):
			normal=mult_s3(-1,normal,buff1)
		faces[i].normal=Math.Vector3(normal)
		cog=mult_s3(1./3, add3(add3(points[0],points[1],buff1),points[2],buff2),buff3)
		faces[i].cog=Math.Vector3(cog)
	
	



"""
Copyright (c) 2017, 2018, 2019 Irstea
Copyright (c) 2017, 2018, 2019 François Kneib
Copyright (c) 2017, 2018, 2019 Franck Bourrier
Copyright (c) 2017, 2018, 2019 David Toe
Copyright (c) 2017, 2018, 2019 Frédéric Berger
Copyright (c) 2017, 2018, 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""

#cython: language_level=3, boundscheck=False, wraparound=False

import numpy as np
import quaternion
from libc.math cimport sin,cos,sqrt
from scipy.stats import gamma

class Vector1(np.ndarray):	#this code derives from the numpy doc for subclassing np.ndarray.
	def __new__(subtype, input=0.0):
		assert(type(input)==type(0.0) or len(input)==1)
		obj = super(Vector1, subtype).__new__(subtype, 1, float,None,0,None,None)
		if(type(input)==type(0.0)):
			obj[0]=input
			return obj
		else:
			obj[0]=input[0]
			return obj
	
	def dot(double[:] A,double[:] B):
		return A[0]*B[0]
		
	def norm(double[:] A):
		return sqrt(A[0]**2)
	
	def normalized(double[:] A):
			cdef double norm = sqrt(A[0]**2)
			return Vector1([A[0]/norm])
	
#NOTE: bind the Vector1 methods outside the class so that they can be used by PlatRock without instanciating a Vector1 object.
dot1=Vector1.dot
normalized1=Vector1.normalized
norm1=Vector1.norm

class Vector2(np.ndarray):	#this code derives from the numpy doc for subclassing np.ndarray.
	def __new__(subtype, input=[0,0]):
		assert(len(input)==2)
		obj = super(Vector2, subtype).__new__(subtype, 2, float,None,0,None,None)
		obj[0]=input[0]
		obj[1]=input[1]
		return obj
	
	def dot(double[:] A,double[:] B):
		return A[0]*B[0]+A[1]*B[1]
		
	def cross(double[:] A,double[:] B):
		return Vector1(A[0]*B[1]-A[1]*B[0])

	def norm(double[:] A):
		return sqrt(A[0]**2+A[1]**2)
	
	def normalized(double[:] A):
		#try:
			cdef double norm = sqrt(A[0]**2+A[1]**2)
			return Vector2([A[0]/norm,A[1]/norm])
		#except:
			#return A*0.

	def rotated(double[:] A,double ang,ax=None):
		cdef double c=cos(ang)
		cdef double s=sin(ang)
		return Vector2([A[0]*c - A[1]*s , A[0]*s + A[1]*c])
	
#NOTE: bind the Vector2 methods outside the class so that they can be used by PlatRock without instanciating a Vector2 object.
cross2=Vector2.cross
dot2=Vector2.dot
normalized2=Vector2.normalized
rotated2=Vector2.rotated
norm2=Vector2.norm

class Vector3(np.ndarray):	#this code derives from the numpy doc for subclassing np.ndarray.
	def __new__(subtype, input=[0,0,0]):
		assert(len(input)==3)
		obj = super(Vector3, subtype).__new__(subtype, 3, float,None,0,None,None)
		obj[0]=input[0]
		obj[1]=input[1]
		obj[2]=input[2]
		return obj
	
	def normalized(A):
		try:
			return A/norm3(A)
		except:
			return A*0.
	
	def norm(A):
		return np.sqrt(A[0]**2+A[1]**2+A[2]**2)
	
	def dot(A,B):
		return A[0]*B[0]+A[1]*B[1]+A[2]*B[2]
	
	def cross(A,B):
		return Vector3([A[1]*B[2]-A[2]*B[1],A[2]*B[0]-A[0]*B[2],A[0]*B[1]-B[0]*A[1]])
	
	def rotated(A,ang,ax):
		output_xyz=Vector3()
		c=np.cos(ang)
		s=np.sin(ang)
		ax=ax.normalized()
		output_xyz[0]=A[0]*( ax[0]**2*(1.-c)+c )             +     A[1]*( ax[0]*ax[1]*(1-c)-ax[2]*s )    +     A[2]*( ax[0]*ax[1]*(1-c)+ax[1]*s )
		output_xyz[1]=A[0]*( ax[0]*ax[1]*(1-c)+ax[2]*s )     +     A[1]*( ax[1]**2*(1-c)+c )             +     A[2]*( ax[1]*ax[2]*(1-c)-ax[0]*s )
		output_xyz[2]=A[0]*( ax[0]*ax[2]*(1-c)-ax[1]*s )     +     A[1]*( ax[1]*ax[2]*(1-c)+ax[0]*s )    +     A[2]*( ax[2]**2*(1-c)+c )
		return output_xyz

#Binds:
cross3=Vector3.cross
dot3=Vector3.dot
normalized3=Vector3.normalized
rotated3=Vector3.rotated
norm3=Vector3.norm


#For some reason, if the following isn't done, jsonpickle will load Vector2 and Vector3 as np.ndarray. (saving is okay by default)
import jsonpickle.handlers
import jsonpickle.ext.numpy
class VectorJsonPickeHandler(jsonpickle.ext.numpy.NumpyNDArrayHandler):
	def __init__(self, *args,**kwargs):
		super(VectorJsonPickeHandler,self).__init__(*args,**kwargs)
	def restore(self,data):
		obj=super(VectorJsonPickeHandler,self).restore(data)
		if(len(obj)==1):
			return Vector1(obj)
		elif(len(obj)==2):
			return Vector2(obj)
		elif(len(obj)==3):
			return Vector3(obj)
jsonpickle.handlers.register(Vector1, VectorJsonPickeHandler)
jsonpickle.handlers.register(Vector2, VectorJsonPickeHandler)
jsonpickle.handlers.register(Vector3, VectorJsonPickeHandler)


def rotate_vector(q,A):
	return A + 2.*cross3(q.vec,(q.w*A+cross3(q.vec,A)))/(q.w**2+q.x**2+q.y**2+q.z**2)
quaternion.rotate_vector=rotate_vector


def atan2_signed(y,x):
	if(not y==0.):
		if(x>0.):
			return np.sign(y)*np.arctan(abs(y/x))
		elif(x<0.):
			return np.sign(y)*(np.pi-np.arctan(abs(y/x)))
		else:
			return np.sign(y)*np.pi/2.
	else:
		if(x>0.):
			return 0
		elif(x<0.):
			return np.pi
		else:
			return np.NaN

def atan2_unsigned(y,x):
	value=atan2_signed(y,x)
	if(value<0.):value+=2.*np.pi
	return value

def get_random_value_from_gamma_distribution(mean,std,rand_gen=None):
	return gamma.rvs((mean/std)**2,0,std**2/mean,size=None,random_state=rand_gen)

#from http://math.15873.pagesperso-orange.fr/page9.htm
def get_2D_polygon_center_of_mass(points):
	area=Gx=Gy=0
	for i in range(-1,len(points)-1,1):
		interm=points[i][0]*points[i+1][1]-points[i+1][0]*points[i][1]
		area+=interm
		Gx+=(points[i][0]+points[i+1][0])*interm
		Gy+=(points[i][1]+points[i+1][1])*interm
	area*=0.5
	Gx/=6*area
	Gy/=6*area
	return [Gx,Gy]
	
def get_2D_polygon_mass_inertia(points,density,cog_centered=False):
	if(not cog_centered):
		#Get the COG
		cog = get_2D_polygon_center_of_mass(points)
		#Center the polygon to the COG (so set COG to (0,0))
		centered_points = points[:] - cog
	else:
		centered_points=points
	#Loop on triangles around the COG (0,0) and compute its inertia
	I=0 ; m=0
	for i in range(-1,centered_points.shape[0]-1):
		#A==(0,0)
		B=Vector2(centered_points[i])
		C=Vector2(centered_points[i+1])
		tri_area=(B.cross(C)).norm()/2
		tri_inertia=density*tri_area/6 * ( B.dot(B) + B.dot(C) + C.dot(C) )
		m+=tri_area
		I+=tri_inertia
	return [m*density,I]

def get_2d_polygon_cog_mass_inertia(points,density):
	cog=get_2D_polygon_center_of_mass(points)
	return [cog]+get_2D_polygon_mass_inertia(points, cog, density)

def sort_2d_polygon_vertices(points):
	center=points.mean(axis=0)
	points[:]=points-center
	atan2_list=np.asarray([atan2_unsigned(p[1],p[0]) for p in points])
	points[:]=points[atan2_list.argsort()]+center
	
	
def center_2d_polygon_vertices(points):
	points[:]-=get_2D_polygon_center_of_mass(points)

def rotate_points_around_origin(points,radians):
	pointsX = points[:,0]*np.cos(radians) - points[:,1]*np.sin(radians)
	pointsY = points[:,0]*np.sin(radians) + points[:,1]*np.cos(radians)
	points[:] = np.array([pointsX,pointsY]).transpose()
	
def get_random_convex_polygon(n,Lx,Ly):
	
	#This is Valtr's algorithm, found here : https://cglab.ca/~sander/misc/ConvexGeneration/convex.html
	xPool = np.sort(np.random.rand(n))
	yPool = np.sort(np.random.rand(n))

	minX=xPool[0]
	maxX=xPool[-1]
	minY=yPool[0]
	maxY=yPool[-1]

	xVec = np.zeros(n)
	yVec = np.zeros(n)
	lastTop = minX
	lastBot = minX
	for i in range(1,n-1):
		x = xPool[i]
		if np.random.random()>0.5 :
			xVec[i-1]=x - lastTop
			lastTop = x
		else:
			xVec[i-1]=lastBot - x
			lastBot = x
	xVec[-2]=maxX - lastTop
	xVec[-1]=lastBot - maxX

	lastLeft = minY
	lastRight = minY
	for i in range(1,n-1):
		y = yPool[i]
		if np.random.random()>0.5 :
			yVec[i-1]=y - lastLeft
			lastLeft = y
		else:
			yVec[i-1]=lastRight - y
			lastRight = y
	yVec[-2]=maxY - lastLeft
	yVec[-1]=lastRight - maxY

	np.random.shuffle(yVec)
	vec=np.array((xVec,yVec)).transpose()
	
	angles=np.arctan2(vec[:,1],vec[:,0])
	vec=vec[angles.argsort()]

	vec[:,0]=np.cumsum(vec[:,0])
	vec[:,1]=np.cumsum(vec[:,1])
	
	xShift=minX - vec[:,0].min() - 0.5
	yShift=minY - vec[:,1].min() - 0.5
	
	vec+=np.array([xShift,yShift])
	
	# End of Valtr's algorithm
	
	#Find the largest point-point distance:
	maxDist=-np.inf
	id1=-1 ; id2=-1
	for i in range(n):
		for j in range(i+1,n):
			d=Vector2(vec[i]-vec[j]).norm()
			if(d>maxDist):
				maxDist=d
				id1=i ; id2=j
	
	#Find the angle of the largest point-point distance:
	long_vect = vec[id1]-vec[id2]
	angle = - np.arctan2(long_vect[1],long_vect[0])
	
	#Rotate the polygon to align its principal axis with X
	rotate_points_around_origin(vec,angle)
	
	#Scale with Lx and Ly:
	DX=vec[:,0].max() - vec[:,0].min()
	DY=vec[:,1].max() - vec[:,1].min()
	vec[:,0] *= Lx/DX
	vec[:,1] *= Ly/DY

	return vec





















"""
Copyright (c) 2017, 2018, 2019 Irstea
Copyright (c) 2017, 2018, 2019 François Kneib
Copyright (c) 2017, 2018, 2019 Franck Bourrier
Copyright (c) 2017, 2018, 2019 David Toe
Copyright (c) 2017, 2018, 2019 Frédéric Berger
Copyright (c) 2017, 2018, 2019 Stéphane Lambert

This file is part of PlatRock.

PlatRock is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

PlatRock is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Foobar. If not, see <http://www.gnu.org/licenses/>.
"""

"""
Definition of the project class
TODO-LIST:
- change the serialization method from pickle to XML or JSON which are more universal and not sensible to the program version upgrades.
"""

from . import Debug
import os,re,glob,fnmatch
import numpy as np
import jsonpickle
import platrock.ThreeD.Simulations
projects_root=""
projects_list=[]

class Project(object):
	def __init__(self):
		self.name=""
		self.date=""
		self.author=""
		self.location=""
		self.context=""
		self.aims=""
		self.path=""
		self.simulations=[]
	
	def write_to_file(self):
		path=self.path+"/PlatRockProject"
		f=open(path,'w',encoding="utf8")
		for k in self.__dict__:
			if k in ["path","simulations_names"]:continue
			f.write('###'+k+'\n')
			f.write(str(self.__dict__[k])+"\n")
		f.close()
	
	def get_loaded_simulation(self,path):
		f=open( path, "r",encoding="utf8")
		r=f.read()
		f.close()
		try:
			s=jsonpickle.decode(r)
			if type(s)==dict :
				Debug.info("Jsonpickle didn't found a corresponding class for",path)
				return s
			if isinstance(s.terrain,platrock.TwoD.Objects.Terrain) : #we didn't save the param_zones as jsonpickle don't handle weird types, reconstruct here.
				s.terrain.set_params_zones()
		except:
			Debug.info("Jsonpickle was unable to load the simulation",path)
			Debug.info('The content of the file was:',r)
			s=None
		return s
	
	def update_simulations_list(self):
		simulations=[]
		sim_dirs=next(os.walk(self.path))[1]
		sim_dirs.sort()
		for sim_dir in sim_dirs:
			sim_dir=self.path+"/"+sim_dir
			sim_file=glob.glob(sim_dir+'/setup.json')
			if(len(sim_file)>0):
				s=self.get_loaded_simulation(sim_file[0])
				if type(s)==dict : continue
				elif(s==None): #the file was not readable for some reason
					for s2 in self.simulations: #search in the existing simulations list in memory
						if(s2.get_full_path()==sim_file[0]):
							s=s2	#if found, keep the one in memory instead of loading
							break
				s.project=self
				simulations.append(s)
		self.simulations=simulations
		
	def reload_simulation(self,simulation):
		s=self.get_loaded_simulation(simulation.get_full_path())
		if(s):
			s.project=self
			for i in range(len(self.simulations)):
				if(self.simulations[i].get_full_path()==s.get_full_path()):
					self.simulations[i]=s
					break
			return s
		else:
			return simulation
	
	def has_simulation_name(self,name):
		if name in [s.name for s in self.simulations]:
			return True
		else:
			return False

def find_by_name(name):
	for p in projects_list:
		if(p.name==name):
			return p

def exists(name):
	if name in [p.name for p in projects_list]:
		return True
	else:
		return False

def update_projects_list():
	global projects_list
	projects_list=[]
	w=os.walk(projects_root)
	for path, dirs, files in w:
		for filename in files:
			if(filename=="PlatRockProject"):
				f=open(path+"/"+filename,'r',encoding="utf8")
				filestring=f.read()
				f.close()
				filestring=re.sub(' *\\n','\\n',filestring)
				p=Project()
				p.path=path
				for argname in ["name","date","author","location","context","aims"]:
					if("###"+argname) in filestring:
						value=filestring.split("###"+argname+"\n")[1].split("\n###")[0]
						value=value.strip()
						p.__dict__[argname]=value
				projects_list.append(p)
